import pytest

"""
Order is important currently:
self._get_packages()
self._create_package()
self._get_package()
self._find_package()
self._copy_package()
self._rename_package()
self._remove_package()
"""


@pytest.fixture()
def create_package(remote, token):
    """
    Adds a "testpackage0" for a test.
    :param remote: See the correesponding fixture.
    :param token: See the correesponding fixture.
    """
    package = remote.new_package(token)
    remote.modify_package(package, "name", "testpackage0", token)
    remote.save_package(package, token)


@pytest.fixture()
def remove_package(remote, token):
    """
    Removes a "testpackage0" for a test.
    :param remote: See the correesponding fixture.
    :param token: See the correesponding fixture.
    """
    yield
    remote.remove_package("testpackage0", token)


@pytest.mark.usefixtures("cobbler_xmlrpc_base")
class TestPackage:

    @pytest.mark.usefixtures("remove_package")
    def test_create_package(self, remote, token):
        """
        Test: create/edit a package object
        """

        packages = remote.get_packages(token)
        package = remote.new_package(token)

        assert remote.modify_package(package, "name", "testpackage0", token)
        assert remote.save_package(package, token)

        new_packages = remote.get_packages(token)
        assert len(new_packages) == len(packages) + 1

    @pytest.mark.usefixtures("create_package", "remove_package")
    def test_get_packages(self, remote, token):
        """
        Test: Get packages
        """

        package = remote.get_packages()

    @pytest.mark.usefixtures("create_package", "remove_package")
    def test_get_package(self, remote):
        """
        Test: Get a package object
        """

        package = remote.get_package("testpackage0")

    @pytest.mark.usefixtures("create_package", "remove_package")
    def test_find_package(self, remote, token):
        """
        Test: find a package object
        """

        result = remote.find_package({"name": "testpackage0"}, token)
        assert result

    @pytest.mark.usefixtures("create_package", "remove_package")
    def test_copy_package(self, remote, token):
        """
        Test: copy a package object
        """
        # Arrange --> Done in fixture

        # Act
        package = remote.get_item_handle("package", "testpackage0", token)
        result = remote.copy_package(package, "testpackagecopy", token)

        # Assert
        assert result

        # Cleanup
        remote.remove_package("testpackagecopy", token)

    @pytest.mark.usefixtures("create_package", "remove_package")
    def test_rename_package(self, remote, token):
        """
        Test: rename a package object
        """

        package = remote.get_item_handle("package", "testpackage0", token)
        assert remote.rename_package(package, "testpackage1", token)
        package = remote.get_item_handle("package", "testpackage1", token)
        assert remote.rename_package(package, "testpackage0", token)

    @pytest.mark.usefixtures("create_package")
    def test_remove_package(self, remote, token):
        """
        Test: remove a package object
        """
        # Arrange --> Done in Fixture
        # Act & Assert
        assert remote.remove_package("testpackage0", token)



"""
Copyright 2006-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""
import uuid
from typing import Union

from cobbler import autoinstall_manager, enums, utils, validate
from cobbler.cexceptions import CX
from cobbler.items import item


class Image(item.Item):
    """
    A Cobbler Image. Tracks a virtual or physical image, as opposed to a answer file (autoinst) led installation.
    """

    TYPE_NAME = "image"
    COLLECTION_TYPE = "image"

    def __init__(self, api, *args, **kwargs):
        """
        Constructor

        :param api: The Cobbler API object which is used for resolving information.
        :param args: The arguments which should be passed additionally to the base Item class constructor.
        :param kwargs: The keyword arguments which should be passed additionally to the base Item class constructor.
        """
        super().__init__(api, *args, **kwargs)
        self._arch = enums.Archs.X86_64
        self._autoinstall = enums.VALUE_INHERITED
        self._breed = ""
        self._file = ""
        self._image_type = enums.ImageTypes.DIRECT
        self._network_count = 0
        self._os_version = ""
        self._boot_loaders = []
        self._menu = ""
        self._virt_auto_boot = False
        self._virt_bridge = ""
        self._virt_cpus = 0
        self._virt_disk_driver = enums.VirtDiskDrivers.RAW
        self._virt_file_size = 0.0
        self._virt_path = ""
        self._virt_ram = 0
        self._virt_type = enums.VirtType.AUTO
        self._supported_boot_loaders = []

    def __getattr__(self, name):
        if name == "kickstart":
            return self.autoinstall
        raise AttributeError("Attribute \"%s\" did not exist on object type Image." % name)

    #
    # override some base class methods first (item.Item)
    #

    def make_clone(self):
        """
        Clone this image object. Please manually adjust all value yourself to make the cloned object unique.

        :return: The cloned instance of this object.
        """
        _dict = self.to_dict()
        cloned = Image(self.api)
        cloned.from_dict(_dict)
        cloned.uid = uuid.uuid4().hex
        return cloned

    def from_dict(self, dictionary: dict):
        """
        Initializes the object with attributes from the dictionary.

        :param dictionary: The dictionary with values.
        """
        if "name" in dictionary:
            self.name = dictionary["name"]
        if "parent" in dictionary:
            self.parent = dictionary["parent"]
        self._remove_depreacted_dict_keys(dictionary)
        super().from_dict(dictionary)

    #
    # specific methods for item.Image
    #

    @property
    def arch(self) -> enums.Archs:
        """
        Represents the architecture the image has. If deployed to a physical host this should be enforced, a virtual
        image may be deployed on a host with any architecture.

        :getter: The current architecture. Default is ``X86_64``.
        :setter: Should be of the enum type or str. May raise an exception in case the architecture is not known to
                 Cobbler.
        """
        return self._arch

    @arch.setter
    def arch(self, arch: Union[str, enums.Archs]):
        """
        The field is mainly relevant to PXE provisioning.
        See comments for arch property in distro.py, this works the same.

        :param arch: The new architecture to set.
        """
        self._arch = validate.validate_arch(arch)

    @property
    def autoinstall(self) -> str:
        """
        Property for the automatic installation file path, this must be a local file.

        It may not make sense for images to have automatic installation templates. It really doesn't. However if the
        image type is 'iso' koan can create a virtual floppy and shove an answer file on it, to script an installation.
        This may not be a automatic installation template per se, it might be a Windows answer file (SIF) etc.

        This property can inherit from a parent. Which is actually the default value.

        :getter: The path relative to the template directory.
        :setter: The location of the template relative to the template base directory.
        """
        return self._autoinstall

    @autoinstall.setter
    def autoinstall(self, autoinstall: str):
        """
        Set the automatic installation file path, this must be a local file.

        :param autoinstall: local automatic installation template file path
        """
        autoinstall_mgr = autoinstall_manager.AutoInstallationManager(self.api._collection_mgr)
        self._autoinstall = autoinstall_mgr.validate_autoinstall_template_file_path(autoinstall)

    @property
    def file(self) -> str:
        """
        Stores the image location. This should be accessible on all nodes that need to access it.

        Format: can be one of the following:
        * username:password@hostname:/path/to/the/filename.ext
        * username@hostname:/path/to/the/filename.ext
        * hostname:/path/to/the/filename.ext
        * /path/to/the/filename.ext

        :getter: The path to the image location or an emtpy string.
        :setter: May raise a TypeError or SyntaxError in case the validation of the location fails.
        """
        return self._file

    @file.setter
    def file(self, filename: str):
        """
        The setter for the image location.

        :param filename: The location where the image is stored.
        :raises SyntaxError: In case a protocol was found.
        """
        if not isinstance(filename, str):
            raise TypeError("file must be of type str to be parsable.")

        if not filename:
            self._file = ""
            return

        # validate file location format
        if filename.find("://") != -1:
            raise SyntaxError("Invalid image file path location, it should not contain a protocol")
        uri = filename
        auth = ""
        hostname = ""
        path = ""

        if filename.find("@") != -1:
            auth, filename = filename.split("@")
        # extract the hostname
        # 1. if we have a colon, then everything before it is a hostname
        # 2. if we don't have a colon, there is no hostname
        if filename.find(":") != -1:
            hostname, filename = filename.split(":")
        elif filename[0] != '/':
            raise SyntaxError("invalid file: %s" % filename)
        # raise an exception if we don't have a valid path
        if len(filename) > 0 and filename[0] != '/':
            raise SyntaxError("file contains an invalid path: %s" % filename)
        if filename.find("/") != -1:
            path, filename = filename.rsplit("/", 1)

        if len(filename) == 0:
            raise SyntaxError("missing filename")
        if len(auth) > 0 and len(hostname) == 0:
            raise SyntaxError("a hostname must be specified with authentication details")

    @property
    def os_version(self) -> str:
        r"""
        The operating system version which the image contains.

        :getter: The sanitized operating system version.
        :setter: Accepts a str which will be validated against the ``distro_signatures.json``.
        """
        return self._os_version

    @os_version.setter
    def os_version(self, os_version):
        """
        Set the operating system version with this setter.

        :param os_version: This must be a valid OS-Version.
        """
        self._os_version = validate.validate_os_version(os_version, self.breed)

    @property
    def breed(self) -> str:
        r"""
        The operating system breed.

        :getter: Returns the current breed.
        :setter: When setting this it is validated against the ``distro_signatures.json`` file.
        """
        return self._breed

    @breed.setter
    def breed(self, breed: str):
        """
        Set the operating system breed with this setter.

        :param breed: The breed of the operating system which is available in the image.
        """
        self._breed = validate.validate_breed(breed)

    @property
    def image_type(self) -> enums.ImageTypes:
        """
        Indicates what type of image this is.
        direct     = something like "memdisk", physical only
        iso        = a bootable ISO that pxe's or can be used for virt installs, virtual only
        virt-clone = a cloned virtual disk (FIXME: not yet supported), virtual only
        memdisk    = hdd image (physical only)

        :getter: The enum type value of the image type.
        :setter: Accepts str like and enum type values and raises a TypeError or ValueError in the case of a problem.
        """
        return self._image_type

    @image_type.setter
    def image_type(self, image_type: Union[enums.ImageTypes, str]):
        """
        The setter which accepts enum type or str type values. Latter ones will be automatically converted if possible.

        :param image_type: One of the four options from above.
        :raises TypeError: In case a disallowed type was found.
        :raises ValueError: In case the conversion from str could not successfully executed.
        """
        if not isinstance(image_type, (enums.ImageTypes, str)):
            raise TypeError("image_type must be of type str or enum.ImageTypes")
        if isinstance(image_type, str):
            if not image_type:
                # FIXME: Add None Image type
                self._image_type = enums.ImageTypes.DIRECT
            try:
                image_type = enums.ImageTypes[image_type.upper()]
            except KeyError as error:
                raise ValueError("image_type choices include: %s" % list(map(str, enums.ImageTypes))) from error
        # str was converted now it must be an enum.ImageTypes
        if not isinstance(image_type, enums.ImageTypes):
            raise TypeError("image_type needs to be of type enums.ImageTypes")
        if image_type not in enums.ImageTypes:
            raise ValueError("image type must be one of the following: %s"
                             % ", ".join(list(map(str, enums.ImageTypes))))
        self._image_type = image_type

    @property
    def virt_cpus(self) -> int:
        """
        The amount of vCPU cores used in case the image is being deployed on top of a VM host.

        :getter: The cores used.
        :setter: The new number of cores.
        """
        return self._virt_cpus

    @virt_cpus.setter
    def virt_cpus(self, num: int):
        """
        Setter for the number of virtual cpus.

        :param num: The number of virtual cpu cores.
        """
        self._virt_cpus = validate.validate_virt_cpus(num)

    @property
    def network_count(self) -> int:
        """
        Represents the number of virtual NICs this image has.

        .. deprecated:: 3.3.0
           This is nowhere used in the project and will be removed in a future release.

        :getter: The number of networks.
        :setter: Raises a ``TypeError`` in case the value is not an int.
        """
        return self._network_count

    @network_count.setter
    def network_count(self, network_count: int):
        """
        Setter for the number of networks.

        :param network_count: If None or emtpy will be set to ``1``, otherwise the given integer value will be set.
        :raises TypeError: In case the network_count was not of type int.
        """
        if network_count is None or network_count == "":
            network_count = 1
        if not isinstance(network_count, int):
            raise TypeError("Field network_count of object image needs to be of type int.")
        self._network_count = network_count

    @property
    def virt_auto_boot(self) -> bool:
        r"""
        Whether the VM should be booted when booting the host or not.

        :getter: ``True`` means autoboot is enabled, otherwise VM is not booted automatically.
        :setter: The new state for the property.
        """
        return self._virt_auto_boot

    @virt_auto_boot.setter
    def virt_auto_boot(self, num: bool):
        """
        Setter for the virtual automatic boot option.

        :param num: May be "0" (disabled) or "1" (enabled), will be converted to a real bool.
        """
        self._virt_auto_boot = validate.validate_virt_auto_boot(num)

    @property
    def virt_file_size(self) -> float:
        r"""
        The size of the image and thus the usable size for the guest.

        .. warning:: There is a regression which makes the usage of multiple disks not possible right now. This will be
                     fixed in a future release.

        :getter: The size of the image(s) in GB.
        :setter: The float with the new size in GB.
        """
        return self._virt_file_size

    @virt_file_size.setter
    def virt_file_size(self, num: float):
        """
        Setter for the virtual file size of the image.

        :param num: Is a non-negative integer (0 means default). Can also be a comma seperated list -- for usage with
                    multiple disks
        """
        self._virt_file_size = validate.validate_virt_file_size(num)

    @property
    def virt_disk_driver(self) -> enums.VirtDiskDrivers:
        """
        The type of disk driver used for storing the image.

        :getter: The enum type representation of the disk driver.
        :setter: May be a ``str`` with the name of the disk driver or from the enum type directly.
        """
        return self._virt_disk_driver

    @virt_disk_driver.setter
    def virt_disk_driver(self, driver: enums.VirtDiskDrivers):
        """
        Setter for the virtual disk driver.

        :param driver: The virtual disk driver which will be set.
        """
        self._virt_disk_driver = validate.validate_virt_disk_driver(driver)

    @property
    def virt_ram(self) -> int:
        """
        The amount of RAM given to the guest in MB.

        :getter: The amount of RAM currently assigned to the image.
        :setter: The new amount of ram. Must be an integer.
        """
        return self._virt_ram

    @virt_ram.setter
    def virt_ram(self, num: int):
        """
        Setter for the amount of virtual RAM the machine will have.

        :param num: 0 tells Koan to just choose a reasonable default.
        """
        self._virt_ram = validate.validate_virt_ram(num)

    @property
    def virt_type(self) -> enums.VirtType:
        """
        The type of image used.

        :getter: The value of the virtual machine.
        :setter: May be of the enum type or a str which is then converted to the enum type.
        """
        return self._virt_type

    @virt_type.setter
    def virt_type(self, vtype: enums.VirtType):
        """
        Setter for the virtual type

        :param vtype: May be one of "qemu", "kvm", "xenpv", "xenfv", "vmware", "vmwarew", "openvz" or "auto".
        """
        self._virt_type = validate.validate_virt_type(vtype)

    @property
    def virt_bridge(self) -> str:
        r"""
        The name of the virtual bridge used for networking.

        .. warning:: The new validation for the setter is not working. Thus the inheritance from the settings is broken.

        :getter: The name of the bridge.
        :setter: The new name of the bridge. If set to an empty ``str``, it will be taken from the settings.
        """
        return self._virt_bridge

    @virt_bridge.setter
    def virt_bridge(self, vbridge: str):
        """
        Setter for the virtual bridge which is used.

        :param vbridge: The name of the virtual bridge to use.
        """
        self._virt_bridge = validate.validate_virt_bridge(vbridge)

    @property
    def virt_path(self) -> str:
        """
        Represents the location where the image for the VM is stored.

        :getter: The path.
        :setter: Is being validated for being a reasonable path. If yes is set, otherwise ignored.
        """
        return self._virt_path

    @virt_path.setter
    def virt_path(self, path: str):
        """
        Setter for the virtual path which is used.

        :param path: The path to where the virtual image is stored.
        """
        self._virt_path = validate.validate_virt_path(path)

    @property
    def menu(self) -> str:
        """
        Property to represent the menu which this image should be put into.

        :getter: The name of the menu or an emtpy str.
        :setter: Should only be the name of the menu not the object. May raise ``CX`` in case the menu does not exist.
        """
        return self._menu

    @menu.setter
    def menu(self, menu: str):
        """
        Setter for the menu property.

        :param menu: The menu for the image.
        :raises CX: In case the menu to be set could not be found.
        """
        if menu and menu != "":
            menu_list = self.api.menus()
            if not menu_list.find(name=menu):
                raise CX("menu %s not found" % menu)
        self._menu = menu

    @property
    def supported_boot_loaders(self):
        """
        Read only property which represents the subset of settable bootloaders.

        :getter: The bootloaders which are available for being set.
        """
        try:
            # If we have already loaded the supported boot loaders from the signature, use that data
            return self._supported_boot_loaders
        except:
            # otherwise, refresh from the signatures / defaults
            self._supported_boot_loaders = utils.get_supported_distro_boot_loaders(self)
            return self._supported_boot_loaders

    @property
    def boot_loaders(self) -> list:
        """
        Represents the boot loaders which are able to boot this image.

        :getter: The bootloaders. May be an emtpy list.
        :setter: A list with the supported boot loaders for this image.
        """
        if self._boot_loaders == enums.VALUE_INHERITED:
            return self.supported_boot_loaders
        return self._boot_loaders

    @boot_loaders.setter
    def boot_loaders(self, boot_loaders: list):
        """
        Setter of the boot loaders.

        :param boot_loaders: The boot loaders for the image.
        :raises TypeError: In case this was of a not allowed type.
        :raises ValueError: In case the str which contained the list could not be successfully split.
        """
        # allow the magic inherit string to persist
        if boot_loaders == enums.VALUE_INHERITED:
            self._boot_loaders = enums.VALUE_INHERITED
            return

        if boot_loaders:
            boot_loaders_split = utils.input_string_or_list(boot_loaders)

            if not isinstance(boot_loaders_split, list):
                raise TypeError("boot_loaders needs to be of type list!")

            if not set(boot_loaders_split).issubset(self.supported_boot_loaders):
                raise ValueError("Error with image %s - not all boot_loaders %s are supported %s" %
                                 (self.name, boot_loaders_split, self.supported_boot_loaders))
            self._boot_loaders = boot_loaders_split
        else:
            self._boot_loaders = []



import pytest


@pytest.mark.usefixtures("cobbler_xmlrpc_base")
class TestBackground:
    """
    Class to test various background jobs
    """

    def test_background_acletup(self, remote, token):
        # Arrange

        # Act
        result = remote.background_aclsetup({}, token)

        # Assert
        assert result

    def test_background_buildiso(self, remote, token):
        # Arrange

        # Act
        result = remote.background_buildiso({}, token)

        # Assert
        assert result

    def test_background_hardlink(self, remote, token):
        # Arrange

        # Act
        result = remote.background_hardlink({}, token)

        # Assert
        assert result

    def test_background_import(self, remote, token):
        # Arrange

        # Act
        result = remote.background_import({}, token)

        # Assert
        assert result

    def test_background_replicate(self, remote, token):
        # Arrange

        # Act
        result = remote.background_replicate({}, token)

        # Assert
        assert result

    def test_background_reposync(self, remote, token):
        # Arrange

        # Act
        result = remote.background_reposync({}, token)

        # Assert
        assert result

    def test_background_validate_autoinstall_files(self, remote, token):
        # Arrange

        # Act
        result = remote.background_validate_autoinstall_files({}, token)

        # Assert
        assert result



"""
Copyright 2006-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""
import logging
import time
import os
import uuid
from threading import Lock
from typing import List, Union

from cobbler import utils
from cobbler.items import package, system, item as item_base, image, profile, repo, mgmtclass, distro, file, menu

from cobbler.cexceptions import CX


class Collection:
    """
    Base class for any serializable list of things.
    """

    def __init__(self, collection_mgr):
        """
        Constructor.

        :param collection_mgr: The collection manager to resolve all information with.
        """
        self.collection_mgr = collection_mgr
        self.listing = {}
        self.api = self.collection_mgr.api
        self.lite_sync = None
        self.lock = Lock()
        self.logger = logging.getLogger()

    def __iter__(self):
        """
        Iterator for the collection. Allows list comprehensions, etc.
        """
        for a in list(self.listing.values()):
            yield a

    def __len__(self):
        """
        Returns size of the collection.
        """
        return len(list(self.listing.values()))

    def factory_produce(self, api, seed_data):
        """
        Must override in subclass. Factory_produce returns an Item object from dict.

        :param api: The API to resolve all information with.
        :param seed_data: Unused Parameter in the base collection.
        """
        raise NotImplementedError()

    def remove(self, name: str, with_delete: bool = True, with_sync: bool = True, with_triggers: bool = True,
               recursive: bool = False):
        """
        Remove an item from collection. This method must be overridden in any subclass.

        :param name: Item Name
        :param with_delete: sync and run triggers
        :param with_sync: sync to server file system
        :param with_triggers: run "on delete" triggers
        :param recursive: recursively delete children
        :returns: NotImplementedError
        """
        raise NotImplementedError("Please implement this in a child class of this class.")

    def get(self, name: str):
        """
        Return object with name in the collection

        :param name: The name of the object to retrieve from the collection.
        :return: The object if it exists. Otherwise None.
        """
        return self.listing.get(name.lower(), None)

    def find(self, name: str = "", return_list: bool = False, no_errors=False,
             **kargs: dict) -> Union[List[item_base.Item], item_base.Item, None]:
        """
        Return first object in the collection that matches all item='value' pairs passed, else return None if no objects
        can be found. When return_list is set, can also return a list.  Empty list would be returned instead of None in
        that case.

        :param name: The object name which should be found.
        :param return_list: If a list should be returned or the first match.
        :param no_errors: If errors which are possibly thrown while searching should be ignored or not.
        :param kargs: If name is present, this is optional, otherwise this dict needs to have at least a key with
                      ``name``. You may specify more keys to finetune the search.
        :return: The first item or a list with all matches.
        :raises ValueError: In case no arguments for searching were specified.
        """
        matches = []

        if name:
            kargs["name"] = name

        kargs = self.__rekey(kargs)

        # no arguments is an error, so we don't return a false match
        if len(kargs) == 0:
            raise ValueError("calling find with no arguments")

        # performance: if the only key is name we can skip the whole loop
        if len(kargs) == 1 and "name" in kargs and not return_list:
            try:
                return self.listing.get(kargs["name"].lower(), None)
            except:
                return self.listing.get(kargs["name"], None)

        self.lock.acquire()
        try:
            for (name, obj) in list(self.listing.items()):
                if obj.find_match(kargs, no_errors=no_errors):
                    matches.append(obj)
        finally:
            self.lock.release()

        if not return_list:
            if len(matches) == 0:
                return None
            return matches[0]
        else:
            return matches

    SEARCH_REKEY = {
        'kopts': 'kernel_options',
        'kopts_post': 'kernel_options_post',
        'inherit': 'parent',
        'ip': 'ip_address',
        'mac': 'mac_address',
        'virt-auto-boot': 'virt_auto_boot',
        'virt-file-size': 'virt_file_size',
        'virt-disk-driver': 'virt_disk_driver',
        'virt-ram': 'virt_ram',
        'virt-path': 'virt_path',
        'virt-type': 'virt_type',
        'virt-bridge': 'virt_bridge',
        'virt-cpus': 'virt_cpus',
        'virt-host': 'virt_host',
        'virt-group': 'virt_group',
        'dhcp-tag': 'dhcp_tag',
        'netboot-enabled': 'netboot_enabled',
        'enable_gpxe': 'enable_ipxe',
        'boot_loader': 'boot_loaders',
    }

    def __rekey(self, _dict: dict) -> dict:
        """
        Find calls from the command line ("cobbler system find") don't always match with the keys from the datastructs
        and this makes them both line up without breaking compatibility with either. Thankfully we don't have a LOT to
        remap.

        :param _dict: The dict which should be remapped.
        :return: The dict which can now be understood by the cli.
        """
        new_dict = {}
        for x in list(_dict.keys()):
            if x in self.SEARCH_REKEY:
                newkey = self.SEARCH_REKEY[x]
                new_dict[newkey] = _dict[x]
            else:
                new_dict[x] = _dict[x]
        return new_dict

    def to_list(self) -> list:
        """
        Serialize the collection

        :return: All elements of the collection as a list.
        """
        return [x.to_dict() for x in list(self.listing.values())]

    def from_list(self, _list: list):
        """
        Create all collection object items from ``_list``.

        :param _list: The list with all item dictionaries.
        """
        if _list is None:
            return
        for item_dict in _list:
            item = self.factory_produce(self.api, item_dict)
            self.add(item)

    def copy(self, ref, newname):
        """
        Copy an object with a new name into the same collection.

        :param ref: The reference to the object which should be copied.
        :param newname: The new name for the copied object.
        """
        ref = ref.make_clone()
        ref.uid = uuid.uuid4().hex
        ref.ctime = time.time()
        ref.name = newname
        if ref.COLLECTION_TYPE == "system":
            # this should only happen for systems
            for interface in ref.interfaces:
                # clear all these out to avoid DHCP/DNS conflicts
                ref.interfaces[interface].dns_name = ""
                ref.interfaces[interface].mac_address = ""
                ref.interfaces[interface].ip_address = ""

        self.add(ref, save=True, with_copy=True, with_triggers=True, with_sync=True, check_for_duplicate_names=True,
                 check_for_duplicate_netinfo=False)

    def rename(self, ref: item_base.Item, newname, with_sync: bool = True, with_triggers: bool = True):
        """
        Allows an object "ref" to be given a new name without affecting the rest of the object tree.

        :param ref: The reference to the object which should be renamed.
        :param newname: The new name for the object.
        :param with_sync: If a sync should be triggered when the object is renamed.
        :param with_triggers: If triggers should be run when the object is renamed.
        """
        # Nothing to do when it is the same name
        if newname == ref.name:
            return

        # Save the old name and rename object
        oldname = ref.name
        ref.name = newname

        # for mgmt classes, update all objects that use it
        if ref.COLLECTION_TYPE == "mgmtclass":
            for what in ["distro", "profile", "system"]:
                items = self.api.find_items(what, {"mgmt_classes": oldname})
                for item in items:
                    for i in range(0, len(item.mgmt_classes)):
                        if item.mgmt_classes[i] == oldname:
                            item.mgmt_classes[i] = newname
                    self.api.add_item(what, item, save=True)

        # for menus, update all objects that use it
        if ref.COLLECTION_TYPE == "menu":
            for what in ["profile", "image"]:
                items = self.api.find_items(what, {"menu": oldname})
                for item in items:
                    item.menu = newname
                    self.api.add_item(what, item, save=True)

        # for a repo, rename the mirror directory
        if ref.COLLECTION_TYPE == "repo":
            path = "/var/www/cobbler/repo_mirror/"
            old_path = os.path.join(path, oldname)
            if os.path.exists(old_path):
                new_path = os.path.join(path, ref.name)
                os.renames(old_path, new_path)

        # for a distro, rename the mirror and references to it
        if ref.COLLECTION_TYPE == 'distro':
            path = utils.find_distro_path(self.api.settings(), ref)

            # create a symlink for the new distro name
            utils.link_distro(self.api.settings(), ref)

            # Test to see if the distro path is based directly on the name of the distro. If it is, things need to
            # updated accordingly.
            if os.path.exists(path) and path == str(os.path.join("/var/www/cobbler/distro_mirror/", ref.name)):
                newpath = os.path.join("/var/www/cobbler/distro_mirror/", ref.name)
                os.renames(path, newpath)

                # update any reference to this path ...
                distros = self.api.distros()
                for d in distros:
                    if d.kernel.find(path) == 0:
                        d.kernel = d.kernel.replace(path, newpath)
                        d.initrd = d.initrd.replace(path, newpath)
                        self.collection_mgr.serialize_item(self, d)

        if ref.COLLECTION_TYPE in ('profile', 'system'):
            if ref.parent is not None:
                ref.parent.children.remove(oldname)

        # Now descend to any direct ancestors and point them at the new object allowing the original object to be
        # removed without orphanage. Direct ancestors will either be profiles or systems. Note that we do have to
        # care as setting the parent is only really meaningful for subprofiles. We ideally want a more generic parent
        # setter.
        kids = ref.get_children()
        for k in kids:
            if self.api.find_profile(name=k) is not None:
                k = self.api.find_profile(name=k)
                if k.parent != "":
                    k.parent = newname
                else:
                    k.distro = newname
                self.api.profiles().add(k, save=True, with_sync=with_sync, with_triggers=with_triggers)
            elif self.api.find_menu(name=k) is not None:
                k = self.api.find_menu(name=k)
                k.parent = newname
                self.api.menus().add(k, save=True, with_sync=with_sync, with_triggers=with_triggers)
            elif self.api.find_system(name=k) is not None:
                k = self.api.find_system(name=k)
                k.profile = newname
                self.api.systems().add(k, save=True, with_sync=with_sync, with_triggers=with_triggers)
            else:
                raise CX("Internal error, unknown child type for child \"%s\"!" % k)

        # now delete the old version and add the new one
        self.remove(oldname, with_delete=True, with_triggers=with_triggers)
        self.add(ref, with_triggers=with_triggers, save=True)

    def add(self, ref, save: bool = False, with_copy: bool = False, with_triggers: bool = True, with_sync: bool = True,
            quick_pxe_update: bool = False, check_for_duplicate_names: bool = False,
            check_for_duplicate_netinfo: bool = False):
        """
        Add an object to the collection

        :param ref: The reference to the object.
        :param save: If this is true then the objet is persisted on the disk.
        :param with_copy: Is a bit of a misnomer, but lots of internal add operations can run with "with_copy" as False.
                          True means a real final commit, as if entered from the command line (or basically, by a user).
                          With with_copy as False, the particular add call might just be being run during
                          deserialization, in which case extra semantics around the add don't really apply. So, in that
                          case, don't run any triggers and don't deal with any actual files.
        :param with_sync: If a sync should be triggered when the object is renamed.
        :param with_triggers: If triggers should be run when the object is renamed.
        :param quick_pxe_update: This decides if there should be run a quick or full update after the add was done.
        :param check_for_duplicate_names: If the name of an object should be unique or not.
        :param check_for_duplicate_netinfo: This checks for duplicate network information. This only has an effect on
                                            systems.
        :raises TypError: Raised in case ``ref`` is None.
        :raises ValueError: Raised in case the name of ``ref`` is empty.
        """
        if ref is None:
            raise TypeError("Unable to add a None object")
        if ref.name is None:
            raise ValueError("Unable to add an object without a name")

        ref.check_if_valid()

        if ref.uid == '':
            ref.uid = uuid.uuid4().hex

        if save:
            now = float(time.time())
            if ref.ctime == 0.0:
                ref.ctime = now
            ref.mtime = now

        if self.lite_sync is None:
            self.lite_sync = self.api.get_sync()

        # migration path for old API parameter that I've renamed.
        if with_copy and not save:
            save = with_copy

        if not save:
            # For people that aren't quite aware of the API if not saving the object, you can't run these features.
            with_triggers = False
            with_sync = False

        # Avoid adding objects to the collection if an object of the same/ip/mac already exists.
        self.__duplication_checks(ref, check_for_duplicate_names, check_for_duplicate_netinfo)

        if ref.COLLECTION_TYPE != self.collection_type():
            raise TypeError("API error: storing wrong data type in collection")

        # failure of a pre trigger will prevent the object from being added
        if save and with_triggers:
            utils.run_triggers(self.api, ref, "/var/lib/cobbler/triggers/add/%s/pre/*" % self.collection_type())

        self.lock.acquire()
        try:
            self.listing[ref.name.lower()] = ref
        finally:
            self.lock.release()

        # update children cache in parent object in case it is not in there already
        if ref.parent and ref.name not in ref.parent.children:
            ref.parent.children.append(ref.name)
            self.logger.debug("Added child \"%s\" to parent \"%s\"", ref.name, ref.parent.name)

        # perform filesystem operations
        if save:
            # Save just this item if possible, if not, save the whole collection
            self.collection_mgr.serialize_item(self, ref)

            if with_sync:
                if isinstance(ref, system.System):
                    # we don't need openvz containers to be network bootable
                    if ref.virt_type == "openvz":
                        ref.netboot_enabled = False
                    self.lite_sync.add_single_system(ref.name)
                elif isinstance(ref, profile.Profile):
                    # we don't need openvz containers to be network bootable
                    if ref.virt_type == "openvz":
                        ref.enable_menu = False
                    self.lite_sync.add_single_profile(ref.name)
                elif isinstance(ref, distro.Distro):
                    self.lite_sync.add_single_distro(ref.name)
                elif isinstance(ref, image.Image):
                    self.lite_sync.add_single_image(ref.name)
                elif isinstance(ref, repo.Repo):
                    pass
                elif isinstance(ref, mgmtclass.Mgmtclass):
                    pass
                elif isinstance(ref, package.Package):
                    pass
                elif isinstance(ref, file.File):
                    pass
                elif isinstance(ref, menu.Menu):
                    pass
                else:
                    print("Internal error. Object type not recognized: %s" % type(ref))
            if not with_sync and quick_pxe_update:
                if isinstance(ref, system.System):
                    self.lite_sync.update_system_netboot_status(ref.name)

            # save the tree, so if neccessary, scripts can examine it.
            if with_triggers:
                utils.run_triggers(self.api, ref, "/var/lib/cobbler/triggers/change/*", [])
                utils.run_triggers(self.api, ref, "/var/lib/cobbler/triggers/add/%s/post/*" % self.collection_type(),
                                   [])

    def __duplication_checks(self, ref, check_for_duplicate_names: bool, check_for_duplicate_netinfo: bool):
        """
        Prevents adding objects with the same name. Prevents adding or editing to provide the same IP, or MAC.
        Enforcement is based on whether the API caller requests it.

        :param ref: The refernce to the object.
        :param check_for_duplicate_names: If the name of an object should be unique or not.
        :param check_for_duplicate_netinfo: This checks for duplicate network information. This only has an effect on
                                            systems.
        :raises CX: If a duplicate is found
        """
        # ToDo: Use return bool type to indicate duplicates and only throw CX in real error case.
        # always protect against duplicate names
        if check_for_duplicate_names:
            match = None
            if isinstance(ref, system.System):
                match = self.api.find_system(ref.name)
            elif isinstance(ref, profile.Profile):
                match = self.api.find_profile(ref.name)
            elif isinstance(ref, distro.Distro):
                match = self.api.find_distro(ref.name)
            elif isinstance(ref, repo.Repo):
                match = self.api.find_repo(ref.name)
            elif isinstance(ref, image.Image):
                match = self.api.find_image(ref.name)
            elif isinstance(ref, mgmtclass.Mgmtclass):
                match = self.api.find_mgmtclass(ref.name)
            elif isinstance(ref, package.Package):
                match = self.api.find_package(ref.name)
            elif isinstance(ref, file.File):
                match = self.api.find_file(ref.name)
            elif isinstance(ref, menu.Menu):
                match = self.api.find_menu(ref.name)
            else:
                raise CX("internal error, unknown object type")

            if match:
                raise CX("An object already exists with that name. Try 'edit'?")

        # the duplicate mac/ip checks can be disabled.
        if not check_for_duplicate_netinfo:
            return

        if isinstance(ref, system.System):
            for (name, intf) in list(ref.interfaces.items()):
                match_ip = []
                match_mac = []
                match_hosts = []
                input_mac = intf.mac_address
                input_ip = intf.ip_address
                input_dns = intf.dns_name
                if not self.api.settings().allow_duplicate_macs and input_mac is not None and input_mac != "":
                    match_mac = self.api.find_system(mac_address=input_mac, return_list=True)
                if not self.api.settings().allow_duplicate_ips and input_ip is not None and input_ip != "":
                    match_ip = self.api.find_system(ip_address=input_ip, return_list=True)
                # it's ok to conflict with your own net info.

                if not self.api.settings().allow_duplicate_hostnames and input_dns is not None and input_dns != "":
                    match_hosts = self.api.find_system(dns_name=input_dns, return_list=True)

                for x in match_mac:
                    if x.name != ref.name:
                        raise CX("Can't save system %s. The MAC address (%s) is already used by system %s (%s)"
                                 % (ref.name, intf.mac_address, x.name, name))
                for x in match_ip:
                    if x.name != ref.name:
                        raise CX("Can't save system %s. The IP address (%s) is already used by system %s (%s)"
                                 % (ref.name, intf.ip_address, x.name, name))
                for x in match_hosts:
                    if x.name != ref.name:
                        raise CX("Can't save system %s.  The dns name (%s) is already used by system %s (%s)"
                                 % (ref.name, intf.dns_name, x.name, name))

    def to_string(self) -> str:
        """
        Creates a printable representation of the collection suitable for reading by humans or parsing from scripts.
        Actually scripts would be better off reading the JSON in the cobbler_collections files directly.

        :return: The object as a string representation.
        """
        values = list(self.listing.values())[:]   # copy the values
        values.sort()                       # sort the copy (2.3 fix)
        results = []
        for i, v in enumerate(values):
            results.append(v.to_string())
        if len(values) > 0:
            return "\n\n".join(results)
        else:
            return "No objects found"

    @staticmethod
    def collection_type() -> str:
        """
        Returns the string key for the name of the collection (used by serializer etc)
        """
        raise NotImplementedError("Please implement the method \"collection_type\" in your Collection!")

    @staticmethod
    def collection_types() -> str:
        """
        Returns the string key for the plural name of the collection (used by serializer)
        """
        raise NotImplementedError("Please implement the method \"collection_types\" in your Collection!")



from cobbler import utils
from cobbler.modules.authentication import passthru


class TestPassthrough:
    def test_authenticate_negative(self):
        # Arrange & Act
        result = passthru.authenticate("", "", "")

        # Assert
        assert not result

    def test_authenticate(self, monkeypatch):
        # Arrange
        def mockreturn():
            return "testpassword"
        monkeypatch.setattr(utils, "get_shared_secret", mockreturn)

        # Act
        result = passthru.authenticate("", "", "testpassword")

        # Assert
        assert result



"""
Authentication module that defers to Apache and trusts
what Apache trusts.

Copyright 2008-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This software may be freely redistributed under the terms of the GNU
general public license.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA.
"""

from cobbler import utils


def register() -> str:
    """
    The mandatory Cobbler module registration hook.

    :return: Always "authn"
    """
    return "authn"


def authenticate(api_handle, username, password) -> bool:
    """
    Validate a username/password combo. Uses cobbler_auth_helper

    :param api_handle: This parameter is not used currently.
    :param username: This parameter is not used currently.
    :param password: This should be the internal Cobbler secret.
    :return: True if the password is the secret, otherwise false.
    """
    return password == utils.get_shared_secret()



import optparse

import pytest

from cobbler import cli
from unittest.mock import patch, MagicMock


@pytest.mark.parametrize("input_data,expected_result", [
    ("", ""),
    (0, 0),
    (None, "")
])
def test_n2s(input_data, expected_result):
    # Arrange
    # Act
    result = cli.n2s(input_data)

    # Assert
    assert result == expected_result


@pytest.mark.skip("TODO: Does not work yet.")
@pytest.mark.parametrize("input_options,input_k,expected_result", [
    (None, None, None),
    ("--dns", "name", ""),
    ("--dhcp", "name", ""),
    ("--system", "name", "")
])
def test_opt(input_options, input_k, expected_result, mocker):
    # Arrange
    # TODO: Create Mock which replaces n2s
    # mocker.patch.object(cli, "n2s")

    # Act
    print(cli.opt(input_options, input_k))

    # Assert
    # TODO: Assert args for n2s not the function return
    assert False


@pytest.mark.parametrize("input_parser,expected_result", [
    (["--systems=a.b.c"], {'systems': ['a.b.c']}),
    (["--systems=a.b.c,a.d.c"], {'systems': ['a.b.c', 'a.d.c']}),
    (["--systems=t1.example.de"], {'systems': ['t1.example.de']})
])
def test_get_comma_separated_args(input_parser, expected_result):
    # Arrange
    parser_obj = optparse.OptionParser()
    parser_obj.add_option("--systems", dest="systems", type='string', action="callback",
                          callback=cli.get_comma_separated_args)

    # Act
    (options, args) = parser_obj.parse_args(args=input_parser)

    # Assert
    assert expected_result == parser_obj.values.__dict__



"""
Authentication module that uses ldap
Settings in /etc/cobbler/authn_ldap.conf
Choice of authentication module is in /etc/cobbler/modules.conf

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""


import traceback

from cobbler.cexceptions import CX


def register() -> str:
    """
    The mandatory Cobbler module registration hook.

    :return: Always "authn"
    """

    return "authn"


def authenticate(api_handle, username, password) -> bool:
    """
    Validate an LDAP bind, returning whether the authentication was successful or not.

    :param api_handle: The api instance to resolve settings.
    :param username: The username to authenticate.
    :param password: The password to authenticate.
    :return: True if the ldap server authentication was a success, otherwise false.
    :raises CX: Raised in case the LDAP search bind credentials are missing in the settings.
    """

    if not password:
        return False
    import ldap

    server = api_handle.settings().ldap_server
    basedn = api_handle.settings().ldap_base_dn
    port = str(api_handle.settings().ldap_port)
    prefix = api_handle.settings().ldap_search_prefix

    # Support for LDAP client certificates
    tls_cacertfile = api_handle.settings().ldap_tls_cacertfile
    tls_keyfile = api_handle.settings().ldap_tls_keyfile
    tls_certfile = api_handle.settings().ldap_tls_certfile

    # allow multiple servers split by a space
    if server.find(" "):
        servers = server.split()
    else:
        servers = [server]

    # to get ldap working with Active Directory
    ldap.set_option(ldap.OPT_REFERRALS, 0)

    if tls_cacertfile:
        ldap.set_option(ldap.OPT_X_TLS_CACERTFILE, tls_cacertfile)
    if tls_keyfile:
        ldap.set_option(ldap.OPT_X_TLS_KEYFILE, tls_keyfile)
    if tls_certfile:
        ldap.set_option(ldap.OPT_X_TLS_CERTFILE, tls_certfile)

    uri = ""
    for server in servers:
        # form our ldap uri based on connection port
        if port == '389':
            uri += 'ldap://' + server
        elif port == '636':
            uri += 'ldaps://' + server
        else:
            uri += 'ldap://' + "%s:%s" % (server, port)
        uri += ' '

    uri = uri.strip()

    # connect to LDAP host
    dir = ldap.initialize(uri)

    # start_tls if tls is 'on', 'true' or 'yes' and we're not already using old-SSL
    if port != '636':
        if api_handle.settings().ldap_tls:
            try:
                dir.start_tls_s()
            except:
                traceback.print_exc()
                return False

    # if we're not allowed to search anonymously, grok the search bind settings and attempt to bind
    if api_handle.settings().ldap_anonymous_bind:
        searchdn = api_handle.settings().ldap_search_bind_dn
        searchpw = api_handle.settings().ldap_search_passwd

        if searchdn == '' or searchpw == '':
            raise CX("Missing search bind settings")

        try:
            dir.simple_bind_s(searchdn, searchpw)
        except:
            traceback.print_exc()
            return False

    # perform a subtree search in basedn to find the full dn of the user
    # TODO: what if username is a CN?  maybe it goes into the config file as well?
    filter = prefix + username
    result = dir.search_s(basedn, ldap.SCOPE_SUBTREE, filter, [])
    if result:
        for dn, entry in result:
            # username _should_ be unique so we should only have one result ignore entry; we don't need it
            pass
    else:
        return False

    try:
        # attempt to bind as the user
        dir.simple_bind_s(dn, password)
        dir.unbind()
        return True
    except:
        # traceback.print_exc()
        return False



"""
Report from a Cobbler master.
FIXME: reinstante functionality for 2.0

Copyright 2007-2009, Red Hat, Inc and Others
Anderson Silva <ansilva@redhat.com>
Michael DeHaan <michael.dehaan AT gmail>

This software may be freely redistributed under the terms of the GNU
general public license.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA.
"""
import re
from typing import Optional

from cobbler import utils


class Report:

    def __init__(self, collection_mgr):
        """
        Constructor

        :param collection_mgr: The collection manager to hold all information in Cobbler available.
        """
        self.collection_mgr = collection_mgr
        self.settings = collection_mgr.settings()
        self.api = collection_mgr.api
        self.report_type = None
        self.report_what = None
        self.report_name = None
        self.report_fields = None
        self.report_noheaders = None
        self.array_re = re.compile(r'([^[]+)\[([^]]+)\]')

    def fielder(self, structure: dict, fields_list: list):
        """
        Return data from a subset of fields of some item

        :param structure: The item structure to report.
        :param fields_list: The list of fields which should be returned.
        :return: The same item with only the given subset of information.
        """
        item = {}

        for field in fields_list:
            internal = self.array_re.search(field)
            # check if field is primary field
            if field in list(structure.keys()):
                item[field] = structure[field]
            # check if subfield in 'interfaces' field
            elif internal and internal.group(1) in list(structure.keys()):
                outer = internal.group(1)
                inner = internal.group(2)
                if isinstance(structure[outer], dict) and inner in structure[outer]:

                    item[field] = structure[outer][inner]
            elif "interfaces" in list(structure.keys()):
                for device in list(structure['interfaces'].keys()):
                    if field in structure['interfaces'][device]:
                        item[field] = device + ': ' + structure['interfaces'][device][field]
        return item

    def reporting_csv(self, info, order: list, noheaders: bool) -> str:
        """
        Formats data on 'info' for csv output

        :param info: The list of iteratable items for csv output.
        :param order: The list of fields which are available in the csv file.
        :param noheaders: Whether headers are printed to the output or not.
        :return: The string with the csv.
        """
        outputheaders = ''
        outputbody = ''
        sep = ','

        info_count = 0
        for item in info:

            item_count = 0
            for key in order:

                if info_count == 0:
                    outputheaders += str(key) + sep

                if key in list(item.keys()):
                    outputbody += str(item[key]) + sep
                else:
                    outputbody += '-' + sep

                item_count += 1

            info_count += 1
            outputbody += '\n'

        outputheaders += '\n'

        if noheaders:
            outputheaders = ''

        return outputheaders + outputbody

    def reporting_trac(self, info, order: list, noheaders: bool) -> str:
        """
        Formats data on 'info' for trac wiki table output

        :param info: The list of iteratable items for table output.
        :param order: The list of fields which are available in the table file.
        :param noheaders: Whether headers are printed to the output or not.
        :return: The string with the generated table.
        """
        outputheaders = ''
        outputbody = ''
        sep = '||'

        info_count = 0
        for item in info:

            item_count = 0
            for key in order:

                if info_count == 0:
                    outputheaders += sep + str(key)

                if key in list(item.keys()):
                    outputbody += sep + str(item[key])
                else:
                    outputbody += sep + '-'

                item_count = item_count + 1

            info_count = info_count + 1
            outputbody += '||\n'

        outputheaders += '||\n'

        if noheaders:
            outputheaders = ''

        return outputheaders + outputbody

    def reporting_doku(self, info, order: list, noheaders: bool) -> str:
        """
        Formats data on 'info' for doku wiki table output

        :param info: The list of iteratable items for table output.
        :param order: The list of fields which are available in the table file.
        :param noheaders: Whether headers are printed to the output or not.
        :return: The string with the generated table.
        """
        outputheaders = ''
        outputbody = ''
        sep1 = '^'
        sep2 = '|'

        info_count = 0
        for item in info:

            item_count = 0
            for key in order:

                if info_count == 0:
                    outputheaders += sep1 + key

                if key in list(item.keys()):
                    outputbody += sep2 + item[key]
                else:
                    outputbody += sep2 + '-'

                item_count = item_count + 1

            info_count = info_count + 1
            outputbody += sep2 + '\n'

        outputheaders += sep1 + '\n'

        if noheaders:
            outputheaders = ''

        return outputheaders + outputbody

    def reporting_mediawiki(self, info, order: list, noheaders: bool) -> str:
        """
        Formats data on 'info' for mediawiki table output

        :param info: The list of iteratable items for table output.
        :param order: The list of fields which are available in the table file.
        :param noheaders: Whether headers are printed to the output or not.
        :return: The string with the generated table.
        """
        outputheaders = ''
        outputbody = ''
        opentable = '{| border="1"\n'
        closetable = '|}\n'
        sep1 = '||'
        sep2 = '|'
        sep3 = '|-'

        info_count = 0
        for item in info:

            item_count = 0
            for key in order:

                if info_count == 0 and item_count == 0:
                    outputheaders += sep2 + key
                elif info_count == 0:
                    outputheaders += sep1 + key

                if item_count == 0:
                    if key in list(item.keys()):
                        outputbody += sep2 + str(item[key])
                    else:
                        outputbody += sep2 + '-'
                else:
                    if key in list(item.keys()):
                        outputbody += sep1 + str(item[key])
                    else:
                        outputbody += sep1 + '-'

                item_count = item_count + 1

            info_count = info_count + 1
            outputbody += '\n' + sep3 + '\n'

        outputheaders += '\n' + sep3 + '\n'

        if noheaders:
            outputheaders = ''

        return opentable + outputheaders + outputbody + closetable

    def print_formatted_data(self, data, order: list, report_type: str, noheaders: bool):
        """
        Used for picking the correct format to output data as

        :param data: The list of iteratable items for table output.
        :param order: The list of fields which are available in the table file.
        :param noheaders: Whether headers are printed to the output or not.
        :param report_type: The type of report which should be used.
        """
        if report_type == "csv":
            print(self.reporting_csv(data, order, noheaders))
        if report_type == "mediawiki":
            print(self.reporting_mediawiki(data, order, noheaders))
        if report_type == "trac":
            print(self.reporting_trac(data, order, noheaders))
        if report_type == "doku":
            print(self.reporting_doku(data, order, noheaders))

    def reporting_print_sorted(self, collection):
        """
        Prints all objects in a collection sorted by name

        :param collection: The collection to print.
        """
        collection = [x for x in collection]
        collection.sort(key=lambda x: x.name)
        for x in collection:
            print(x.to_string())

    def reporting_list_names2(self, collection, name: str):
        """
        Prints a specific object in a collection.

        :param collection: The collections object to print a collection from.
        :param name: The name of the collection to print.
        """
        obj = collection.get(name)
        if obj is not None:
            print(obj.to_string())

    def reporting_print_all_fields(self, collection, report_name: str, report_type: str, report_noheaders: bool) -> str:
        """
        Prints all fields in a collection as a table given the report type

        :param collection: The collection to report.
        :param report_name: The name of the report.
        :param report_type: The type of report to give.
        :param report_noheaders: Report without the headers. (May be useful for machine parsing)
        :return: A report with all fields included pretty printed or machine readable.
        """
        # per-item hack
        if report_name:
            collection = collection.find(name=report_name)
            if collection:
                collection = [collection]
            else:
                return ""

        collection = [x for x in collection]
        collection.sort(key=lambda x: x.name)
        data = []
        out_order = []
        count = 0
        for x in collection:
            item = {}
            if x.ITEM_TYPE == "settings":
                structure = x.to_dict()
            else:
                structure = x.to_list()

            for (key, value) in list(structure.items()):
                # exception for systems which could have > 1 interface
                if key == "interfaces":
                    for (device, info) in list(value.items()):
                        for (info_header, info_value) in list(info.items()):
                            item[info_header] = str(device) + ': ' + str(info_value)
                            # needs to create order list for print_formatted_fields
                            if count == 0:
                                out_order.append(info_header)
                else:
                    item[key] = value
                    # needs to create order list for print_formatted_fields
                    if count == 0:
                        out_order.append(key)

            count = count + 1

            data.append(item)

        self.print_formatted_data(data=data, order=out_order, report_type=report_type, noheaders=report_noheaders)

    def reporting_print_x_fields(self, collection, report_name: str, report_type: str, report_fields: str,
                                 report_noheaders: bool):
        """
        Prints specific fields in a collection as a table given the report type

        :param collection: The collection to report.
        :param report_name: The name of the report.
        :param report_type: The type of report to give.
        :param report_fields: The fields which should be included in the report.
        :param report_noheaders: Report without the headers. (May be useful for machine parsing)
        """
        # per-item hack
        if report_name:
            collection = collection.find(name=report_name)
            if collection:
                collection = [collection]
            else:
                return

        collection = [x for x in collection]
        collection.sort(key=lambda x: x.name)
        data = []
        fields_list = report_fields.replace(' ', '').split(',')

        for x in collection:
            if x.ITEM_TYPE == "settings":
                structure = x.to_dict()
            else:
                structure = x.to_list()
            item = self.fielder(structure, fields_list)
            data.append(item)

        self.print_formatted_data(data=data, order=fields_list, report_type=report_type, noheaders=report_noheaders)

    # -------------------------------------------------------

    def run(self, report_what: Optional[str] = None, report_name: Optional[str] = None,
            report_type: Optional[str] = None, report_fields: Optional[str] = None,
            report_noheaders: Optional[bool] = None):
        """
        Get remote profiles and distros and sync them locally

        1. Handles original report output
        2. Handles all fields of report outputs as table given a format
        3. Handles specific fields of report outputs as table given a format

        :param report_what: What should be reported. May be "all".
        :param report_name: The name of the report.
        :param report_type: The type of report to give.
        :param report_fields: The fields which should be included in the report.
        :param report_noheaders: Report without the headers. (May be useful for machine parsing)
        """
        if report_type == 'text' and report_fields == 'all':
            for collection_name in ["distro", "profile", "system", "repo", "network", "image", "mgmtclass", "package",
                                    "file"]:
                if report_what == "all" or report_what == collection_name or report_what == "%ss" % collection_name \
                        or report_what == "%ses" % collection_name:
                    if report_name:
                        self.reporting_list_names2(self.api.get_items(collection_name), report_name)
                    else:
                        self.reporting_print_sorted(self.api.get_items(collection_name))

        elif report_type == 'text' and report_fields != 'all':
            utils.die("The 'text' type can only be used with field set to 'all'")

        elif report_type != 'text' and report_fields == 'all':
            for collection_name in ["distro", "profile", "system", "repo", "network", "image", "mgmtclass", "package",
                                    "file"]:
                if report_what == "all" or report_what == collection_name or report_what == "%ss" % collection_name \
                        or report_what == "%ses" % collection_name:
                    self.reporting_print_all_fields(self.api.get_items(collection_name), report_name, report_type,
                                                    report_noheaders)

        else:
            for collection_name in ["distro", "profile", "system", "repo", "network", "image", "mgmtclass", "package",
                                    "file"]:
                if report_what == "all" or report_what == collection_name or report_what == "%ss" % collection_name \
                        or report_what == "%ses" % collection_name:
                    self.reporting_print_x_fields(self.api.get_items(collection_name), report_name, report_type,
                                                  report_fields, report_noheaders)



"""
Copyright 2006-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""

import os.path
import glob

from cobbler.cobbler_collections import collection
from cobbler.items import distro
from cobbler import utils
from cobbler.cexceptions import CX


class Distros(collection.Collection):
    """
    A distro represents a network bootable matched set of kernels and initrd files.
    """

    @staticmethod
    def collection_type() -> str:
        return "distro"

    @staticmethod
    def collection_types() -> str:
        return "distros"

    def factory_produce(self, api, item_dict):
        """
        Return a Distro forged from item_dict
        """
        new_distro = distro.Distro(api)
        new_distro.from_dict(item_dict)
        return new_distro

    def remove(self, name, with_delete: bool = True, with_sync: bool = True, with_triggers: bool = True,
               recursive: bool = False):
        """
        Remove element named 'name' from the collection

        :raises CX: In case any subitem (profiles or systems) would be orphaned. If the option ``recursive`` is set then
                    the orphaned items would be removed automatically.
        """
        name = name.lower()

        # first see if any Groups use this distro
        if not recursive:
            for profile in self.api.profiles():
                if profile.distro and profile.distro.name.lower() == name:
                    raise CX("removal would orphan profile: %s" % profile.name)

        obj = self.find(name=name)

        if obj is not None:
            kernel = obj.kernel
            if recursive:
                kids = obj.get_children()
                for k in kids:
                    self.api.remove_profile(k, recursive=recursive, delete=with_delete, with_triggers=with_triggers)

            if with_delete:
                if with_triggers:
                    utils.run_triggers(self.api, obj, "/var/lib/cobbler/triggers/delete/distro/pre/*", [])
                if with_sync:
                    lite_sync = self.api.get_sync()
                    lite_sync.remove_single_distro(name)
            self.lock.acquire()
            try:
                del self.listing[name]
            finally:
                self.lock.release()

            self.collection_mgr.serialize_delete(self, obj)

            if with_delete:
                if with_triggers:
                    utils.run_triggers(self.api, obj, "/var/lib/cobbler/triggers/delete/distro/post/*", [])
                    utils.run_triggers(self.api, obj, "/var/lib/cobbler/triggers/change/*", [])

            # look through all mirrored directories and find if any directory is holding this particular distribution's
            # kernel and initrd
            settings = self.api.settings()
            possible_storage = glob.glob(settings.webdir + "/distro_mirror/*")
            path = None
            for storage in possible_storage:
                if os.path.dirname(obj.kernel).find(storage) != -1:
                    path = storage
                    continue

            # if we found a mirrored path above, we can delete the mirrored storage /if/ no other object is using the
            # same mirrored storage.
            if with_delete and path is not None and os.path.exists(path) and kernel.find(settings.webdir) != -1:
                # this distro was originally imported so we know we can clean up the associated storage as long as
                # nothing else is also using this storage.
                found = False
                distros = self.api.distros()
                for d in distros:
                    if d.kernel.find(path) != -1:
                        found = True
                if not found:
                    utils.rmtree(path)



import pytest

from cobbler.cexceptions import CX
from cobbler import module_loader
from tests.conftest import does_not_raise


@pytest.fixture(scope="function")
def reset_modules():
    module_loader.MODULE_CACHE = {}
    module_loader.MODULES_BY_CATEGORY = {}


@pytest.fixture(scope="function")
def load_modules():
    module_loader.load_modules()


def test_load_modules():
    # Arrange

    # Act
    module_loader.load_modules()

    # Assert
    assert module_loader.MODULE_CACHE != {}
    assert module_loader.MODULES_BY_CATEGORY != {}


@pytest.mark.usefixtures("reset_modules", "load_modules")
@pytest.mark.parametrize("module_name", [
    ("nsupdate_add_system_post"),
    ("nsupdate_delete_system_pre"),
    ("scm_track"),
    ("sync_post_restart_services")
    # ("sync_post_wingen")
])
def test_get_module_by_name(module_name):
    # Arrange -> Done in fixtures

    # Act
    returned_module = module_loader.get_module_by_name(module_name)

    # Assert
    assert isinstance(returned_module.register(), str)


@pytest.mark.usefixtures("reset_modules", "load_modules")
@pytest.mark.parametrize("module_section,fallback_name,expected_result,expected_exception", [
    ("authentication", "", "authentication.configfile", does_not_raise()),
    ("authorization", "", "authorization.allowall", does_not_raise()),
    ("dns", "", "managers.bind", does_not_raise()),
    ("dhcp", "", "managers.isc", does_not_raise()),
    ("tftpd", "", "managers.in_tftpd", does_not_raise()),
    ("wrong_section", None, "", pytest.raises(CX)),
    ("wrong_section", "authentication.configfile", "authentication.configfile", does_not_raise())
])
def test_get_module_name(module_section, fallback_name, expected_result, expected_exception):
    # Arrange -> Done in fixtures

    # Act
    with expected_exception:
        result_name = module_loader.get_module_name(module_section, "module", fallback_name)

        # Assert
        assert result_name == expected_result


@pytest.mark.usefixtures("reset_modules", "load_modules")
@pytest.mark.parametrize("module_section,fallback_name,expected_exception", [
    ("authentication", "", does_not_raise()),
    ("authorization", "", does_not_raise()),
    ("dns", "", does_not_raise()),
    ("dhcp", "", does_not_raise()),
    ("tftpd", "", does_not_raise()),
    ("wrong_section", "", pytest.raises(CX)),
    ("wrong_section", "authentication.configfile", does_not_raise())
])
def test_get_module_from_file(module_section, fallback_name, expected_exception):
    # Arrange -> Done in fixtures

    # Act
    with expected_exception:
        result_module = module_loader.get_module_from_file(module_section, "module", fallback_name)

        # Assert
        assert isinstance(result_module.register(), str)


@pytest.mark.usefixtures("reset_modules", "load_modules")
@pytest.mark.parametrize("category,expected_names", [
    (None, ["cobbler.modules.sync_post_wingen"]),
    ("/var/lib/cobbler/triggers/add/system/post/*", ["cobbler.modules.nsupdate_add_system_post"]),
    ("/var/lib/cobbler/triggers/sync/post/*", ["cobbler.modules.sync_post_restart_services"]),
    ("/var/lib/cobbler/triggers/delete/system/pre/*", ["cobbler.modules.nsupdate_delete_system_pre"]),
    ("/var/lib/cobbler/triggers/change/*", ["cobbler.modules.managers.genders", "cobbler.modules.scm_track"]),
    ("/var/lib/cobbler/triggers/install/post/*", ["cobbler.modules.installation.post_log",
                                                  "cobbler.modules.installation.post_power",
                                                  "cobbler.modules.installation.post_puppet",
                                                  "cobbler.modules.installation.post_report"]),
    ("/var/lib/cobbler/triggers/install/pre/*", ["cobbler.modules.installation.pre_clear_anamon_logs",
                                                 "cobbler.modules.installation.pre_log",
                                                 "cobbler.modules.installation.pre_puppet"]),
    ("manage", ["cobbler.modules.managers.bind", "cobbler.modules.managers.dnsmasq",
                "cobbler.modules.managers.in_tftpd", "cobbler.modules.managers.isc",
                "cobbler.modules.managers.ndjbdns"]),
    ("manage/import", ["cobbler.modules.managers.import_signatures"]),
    ("serializer", ["cobbler.modules.serializers.file", "cobbler.modules.serializers.mongodb"]),
    ("authz", ["cobbler.modules.authorization.allowall", "cobbler.modules.authorization.configfile",
               "cobbler.modules.authorization.ownership"]),
    ("authn", ["cobbler.modules.authentication.configfile", "cobbler.modules.authentication.denyall",
               "cobbler.modules.authentication.ldap", "cobbler.modules.authentication.pam",
               "cobbler.modules.authentication.passthru", "cobbler.modules.authentication.spacewalk",
               "cobbler.modules.authentication.testing"]),
])
def test_get_modules_in_category(category, expected_names):
    # Arrange -> Done in fixtures

    # Act
    result = module_loader.get_modules_in_category(category)

    # Assert
    assert len(result) > 0
    actual_result = []
    for name in result:
        actual_result.append(name.__name__)
    actual_result.sort()
    assert actual_result == expected_names



import pytest

from cobbler.template_api import CobblerTemplate


class TestCobblerTemplate:
    def test_compile(self):
        # Arrange

        # Act
        compiled_template = CobblerTemplate(searchList=[{"autoinstall_snippets_dir": "/var/lib/cobbler/snippets"}]) \
            .compile(source="$test")
        result = str(compiled_template(namespaces={"test": 5}))

        # Assert
        assert result == "5"

    def test_no_snippets_dir(self):
        # Arrange
        test_template = CobblerTemplate()

        # Act & Assert
        with pytest.raises(AttributeError):
            test_template.read_snippet("nonexistingsnippet")

    def test_read_snippet_none(self):
        # Arrange
        test_template = CobblerTemplate(searchList=[{"autoinstall_snippets_dir": "/var/lib/cobbler/snippets"}])

        # Act
        result = test_template.read_snippet("nonexistingsnippet")

        # Assert
        assert result is None

    def test_read_snippet(self):
        # Arrange
        test_template = CobblerTemplate(searchList=[{"autoinstall_snippets_dir": "/var/lib/cobbler/snippets"}])
        expected = "#errorCatcher ListErrors\n" + "set -x -v\n" + "exec 1>/root/ks-post.log 2>&1\n"

        # Act
        result = test_template.read_snippet("log_ks_post")

        # Assert
        assert result == expected

    def test_nonexisting_snippet(self):
        # Arrange
        test_template = CobblerTemplate(searchList=[{"autoinstall_snippets_dir": "/var/lib/cobbler/snippets"}])

        # Act
        result = test_template.SNIPPET("preseed_early_default")

        # Assert
        assert result == "# Error: no snippet data for preseed_early_default\n"

    def test_snippet(self):
        # Arrange
        test_template = CobblerTemplate(searchList=[{"autoinstall_snippets_dir": "/var/lib/cobbler/snippets"}])

        # Act
        result = test_template.SNIPPET("post_run_deb")

        # Assert
        assert result == "# A general purpose snippet to add late-command actions for preseeds\n"

    def test_sedesc(self):
        # Arrange
        test_input = "This () needs [] to ^ be * escaped {}."
        expected = "This \\(\\) needs \\[\\] to \\^ be \\* escaped \\{\\}\\."
        test_template = CobblerTemplate()

        # Act
        result = test_template.sedesc(test_input)

        # Assert
        assert result == expected



import os

import pytest

dummy_file_path = "/root/dummy"


@pytest.fixture(scope="class")
def setup():
    """
    Initializes testcase
    """
    # create files if necessary
    if not os.path.exists(dummy_file_path):
        open(dummy_file_path, 'w').close()


@pytest.fixture(scope="class")
def teardown():
    """
    Cleans up testcase
    """
    yield
    # remove files
    if os.path.exists(dummy_file_path):
        os.remove(dummy_file_path)


@pytest.fixture(scope="function")
def generate_run_cmd_array():
    def _generate_run_cmd_array(dict_to_convert):
        result_array = []
        for key in dict_to_convert:
            result_array.append("--%s=%s" % (key, dict_to_convert[key]))
        return result_array

    return _generate_run_cmd_array


@pytest.fixture(scope="function")
def add_object_via_cli(run_cmd, generate_run_cmd_array):
    def _add_object_via_cli(object_type, attributes):
        cmd_list = [object_type, "add"]
        options = generate_run_cmd_array(attributes)
        cmd_list.extend(options)
        run_cmd(cmd=cmd_list)

    return _add_object_via_cli


@pytest.fixture(scope="function")
def remove_object_via_cli(run_cmd):
    def _remove_object_via_cli(object_type, name):
        run_cmd(cmd=[object_type, "remove", "--name=%s" % name])

    return _remove_object_via_cli


@pytest.mark.usefixtures("setup", "teardown")
class TestCobblerCliTestObject:
    """
    Test CLI commands on objects
    """

    def test_report(self, run_cmd):
        # Arrange
        expected = """distros:
==========

profiles:
==========

systems:
==========

repos:
==========

images:
==========

mgmtclasses:
==========

packages:
==========

files:
==========

menus:
==========
"""

        # Act
        (outputstd, outputerr) = run_cmd(cmd=["report"])

        # Assert
        assert outputstd == expected

    @pytest.mark.parametrize("object_type", ["distro", "profile", "system", "image", "repo", "package", "mgmtclass",
                                             "file", "menu"])
    def test_report_with_type(self, run_cmd, object_type):
        # Arrange

        # Act
        (outputstd, outputerr) = run_cmd(cmd=[object_type, "report"])

        # Assert
        assert outputstd is None or not outputstd

    @pytest.mark.parametrize("object_type", ["distro", "profile", "system", "image", "repo", "package", "mgmtclass",
                                             "file", "menu"])
    def test_report_with_type_and_name(self, run_cmd, object_type):
        # Arrange
        name = "notexisting"

        # Act
        (outputstd, outputerr) = run_cmd(cmd=[object_type, "report", "--name=%s" % name])

        # Assert
        assert outputstd == "No %s found: %s\n" % (object_type, name)

    @pytest.mark.parametrize("object_type,attributes,to_change,attr_long_name", [
        ("distro",
         {"name": "testdistroedit", "kernel": "Set in method", "initrd": "Set in method", "breed": "suse",
          "arch": "x86_64"},
         ["comment", "Testcomment"], "Comment"),
        ("profile", {"name": "testprofileedit", "distro": "test_distro_edit_profile"},
         ["comment", "Testcomment"], "Comment"),
        ("system", {"name": "test_system_edit", "profile": "test_profile_edit_system"}, ["comment", "Testcomment"],
         "Comment"),
        ("image", {"name": "test_image_edit"}, ["comment", "Testcomment"], "Comment"),
        ("repo", {"name": "testrepoedit", "mirror": "http://localhost"}, ["comment", "Testcomment"], "Comment"),
        ("package", {"name": "testpackageedit"}, ["comment", "Testcomment"], "Comment"),
        ("mgmtclass", {"name": "testmgmtclassedit"}, ["comment", "Testcomment"], "Comment"),
        ("file", {"name": "testfileedit", "path": "/tmp", "owner": "root", "group": "root", "mode": "600",
                  "is-dir": "True"}, ["path", "/test_dir"], "Path"),
        ("menu", {"name": "testmenuedit"}, ["comment", "Testcomment"], "Comment"),
    ])
    def test_edit(self, run_cmd, add_object_via_cli, remove_object_via_cli, create_kernel_initrd, fk_kernel, fk_initrd,
                  object_type, attributes, to_change, attr_long_name):
        # Arrange
        folder = create_kernel_initrd(fk_kernel, fk_initrd)
        kernel_path = os.path.join(folder, fk_kernel)
        initrd_path = os.path.join(folder, fk_kernel)
        name_distro_profile = "test_distro_edit_profile"
        name_distro_system = "test_distro_edit_system"
        name_profile_system = "test_profile_edit_system"
        if object_type == "distro":
            attributes["kernel"] = kernel_path
            attributes["initrd"] = initrd_path
        elif object_type == "profile":
            add_object_via_cli("distro", {"name": name_distro_profile, "kernel": kernel_path, "initrd": initrd_path,
                                          "breed": "suse", "arch": "x86_64"})
        elif object_type == "system":
            add_object_via_cli("distro", {"name": name_distro_system, "kernel": kernel_path, "initrd": initrd_path,
                                          "breed": "suse", "arch": "x86_64"})
            add_object_via_cli("profile", {"name": name_profile_system, "distro": name_distro_system})
        add_object_via_cli(object_type, attributes)

        # Act
        run_cmd(cmd=[object_type, "edit", "--name=%s" % attributes["name"], "--%s='%s'" % (to_change[0], to_change[1])])
        (outputstd, outputerr) = run_cmd(cmd=[object_type, "report", "--name=%s" % attributes["name"]])

        # Cleanup
        remove_object_via_cli(object_type, attributes["name"])
        if object_type == "profile":
            remove_object_via_cli("distro", name_distro_profile)
        elif object_type == "system":
            remove_object_via_cli("profile", name_profile_system)
            remove_object_via_cli("distro", name_distro_system)

        # Assert
        expected = attr_long_name + ":'" + to_change[1] + "'"
        print("Expected: \"" + expected + "\"")
        lines = outputstd.split("\n")
        found = False
        for line in lines:
            line = line.replace(" ", "")
            print("Line: \"" + line + "\"")
            if line == expected:
                found = True
        assert found

    @pytest.mark.parametrize("object_type,attributes", [
        ("distro", {"name": "testdistrofind", "kernel": "Set in method", "initrd": "Set in method", "breed": "suse",
                    "arch": "x86_64"}),
        ("profile", {"name": "testprofilefind", "distro": ""}),
        ("system", {"name": "testsystemfind", "profile": ""}),
        ("image", {"name": "testimagefind"}),
        ("repo", {"name": "testrepofind", "mirror": "http://localhost"}),
        ("package", {"name": "testpackagefind"}),
        ("mgmtclass", {"name": "testmgmtclassfind"}),
        ("file", {"name": "testfilefind", "path": "/tmp", "owner": "root", "group": "root", "mode": "600",
                  "is-dir": "True"}),
        ("menu", {"name": "testmenufind"}),
    ])
    def test_find(self, run_cmd, add_object_via_cli, remove_object_via_cli, create_kernel_initrd, fk_initrd, fk_kernel,
                  object_type, attributes):
        # Arrange
        folder = create_kernel_initrd(fk_kernel, fk_initrd)
        kernel_path = os.path.join(folder, fk_kernel)
        initrd_path = os.path.join(folder, fk_kernel)
        name_distro_profile = "testdistro_find_profile"
        name_distro_system = "testdistro_find_system"
        name_profile_system = "testprofile_find_system"
        if object_type == "distro":
            attributes["kernel"] = kernel_path
            attributes["initrd"] = initrd_path
        elif object_type == "profile":
            attributes["distro"] = name_distro_profile
            add_object_via_cli("distro", {"name": name_distro_profile, "kernel": kernel_path, "initrd": initrd_path,
                                          "breed": "suse", "arch": "x86_64"})
        elif object_type == "system":
            attributes["profile"] = name_profile_system
            add_object_via_cli("distro", {"name": name_distro_system, "kernel": kernel_path, "initrd": initrd_path,
                                          "breed": "suse", "arch": "x86_64"})
            add_object_via_cli("profile", {"name": name_profile_system, "distro": name_distro_system})
        add_object_via_cli(object_type, attributes)

        # Act
        (outputstd, outputerr) = run_cmd(cmd=[object_type, "find", "--name='%s'" % attributes["name"]])

        # Cleanup
        remove_object_via_cli(object_type, attributes["name"])
        if object_type == "profile":
            remove_object_via_cli("distro", name_distro_profile)
        elif object_type == "system":
            remove_object_via_cli("profile", name_profile_system)
            remove_object_via_cli("distro", name_distro_system)

        # Assert
        lines = outputstd.split("\n")
        assert len(lines) >= 1

    @pytest.mark.parametrize("object_type,attributes", [
        ("distro", {"name": "testdistrocopy", "kernel": "Set in method", "initrd": "Set in method", "breed": "suse",
                    "arch": "x86_64"}),
        ("profile", {"name": "testprofilecopy", "distro": "testdistro_copy_profile"}),
        ("system", {"name": "testsystemcopy", "profile": "testprofile_copy_system"}),
        ("image", {"name": "testimagecopy"}),
        ("repo", {"name": "testrepocopy", "mirror": "http://localhost"}),
        ("package", {"name": "testpackagecopy"}),
        ("mgmtclass", {"name": "testmgmtclasscopy"}),
        ("file", {"name": "testfilecopy", "path": "/tmp", "owner": "root", "group": "root", "mode": "600",
                  "is-dir": "True"}),
        ("menu", {"name": "testmenucopy"}),
    ])
    def test_copy(self, run_cmd, add_object_via_cli, remove_object_via_cli, create_kernel_initrd, fk_initrd, fk_kernel,
                  object_type, attributes):
        # Arrange
        folder = create_kernel_initrd(fk_kernel, fk_initrd)
        kernel_path = os.path.join(folder, fk_kernel)
        initrd_path = os.path.join(folder, fk_kernel)
        name_distro_profile = "testdistro_copy_profile"
        name_distro_system = "testdistro_copy_system"
        name_profile_system = "testprofile_copy_system"
        if object_type == "distro":
            attributes["kernel"] = kernel_path
            attributes["initrd"] = initrd_path
        elif object_type == "profile":
            add_object_via_cli("distro", {"name": name_distro_profile, "kernel": kernel_path, "initrd": initrd_path,
                                          "breed": "suse", "arch": "x86_64"})
        elif object_type == "system":
            add_object_via_cli("distro", {"name": name_distro_system,  "kernel": kernel_path, "initrd": initrd_path,
                                          "breed": "suse", "arch": "x86_64"})
            add_object_via_cli("profile", {"name": name_profile_system, "distro": name_distro_system})
        add_object_via_cli(object_type, attributes)
        new_object_name = "%s-copy" % attributes["name"]

        # Act
        (outputstd, outputerr) = run_cmd(cmd=[object_type, "copy", "--name=%s" % attributes["name"], "--newname=%s"
                                              % new_object_name])

        # Cleanup
        remove_object_via_cli(object_type, attributes["name"])
        remove_object_via_cli(object_type, new_object_name)
        if object_type == "profile":
            remove_object_via_cli("distro", name_distro_profile)
        elif object_type == "system":
            remove_object_via_cli("profile", name_profile_system)
            remove_object_via_cli("distro", name_distro_system)

        # Assert
        assert not outputstd

    @pytest.mark.parametrize("object_type,attributes", [
        ("distro", {"name": "testdistrorename", "kernel": "Set in method", "initrd": "Set in method", "breed": "suse",
                    "arch": "x86_64"}),
        ("profile", {"name": "testprofilerename", "distro": "testdistro_rename_profile"}),
        ("system", {"name": "testsystemrename", "profile": "testprofile_rename_system"}),
        ("image", {"name": "testimagerename"}),
        ("repo", {"name": "testreporename", "mirror": "http://localhost"}),
        ("package", {"name": "testpackagerename"}),
        ("mgmtclass", {"name": "testmgmtclassrename"}),
        ("file", {"name": "testfilerename", "path": "/tmp", "owner": "root", "group": "root", "mode": "600",
                  "is-dir": "True"}),
        ("menu", {"name": "testmenurename"}),
    ])
    def test_rename(self, run_cmd, add_object_via_cli, remove_object_via_cli, create_kernel_initrd, fk_initrd,
                    fk_kernel, object_type, attributes):
        # Arrange
        folder = create_kernel_initrd(fk_kernel, fk_initrd)
        kernel_path = os.path.join(folder, fk_kernel)
        initrd_path = os.path.join(folder, fk_kernel)
        name_distro_profile = "testdistro_rename_profile"
        name_distro_system = "testdistro_rename_system"
        name_profile_system = "testprofile_rename_system"
        if object_type == "distro":
            attributes["kernel"] = kernel_path
            attributes["initrd"] = initrd_path
        elif object_type == "profile":
            add_object_via_cli("distro", {"name": name_distro_profile, "kernel": kernel_path, "initrd": initrd_path,
                                          "breed": "suse", "arch": "x86_64"})
        elif object_type == "system":
            add_object_via_cli("distro", {"name": name_distro_system, "kernel": kernel_path, "initrd": initrd_path,
                                          "breed": "suse", "arch": "x86_64"})
            add_object_via_cli("profile", {"name": name_profile_system, "distro": name_distro_system})
        add_object_via_cli(object_type, attributes)
        new_object_name = "%s-renamed" % attributes["name"]

        # Act
        (outputstd, outputerr) = run_cmd(
            cmd=[object_type, "rename", "--name=%s" % attributes["name"], "--newname=%s" % new_object_name])

        # Cleanup
        remove_object_via_cli(object_type, new_object_name)
        if object_type == "profile":
            remove_object_via_cli("distro", name_distro_profile)
        elif object_type == "system":
            remove_object_via_cli("profile", name_profile_system)
            remove_object_via_cli("distro", name_distro_system)

        # Assert
        assert not outputstd

    @pytest.mark.parametrize("object_type,attributes", [
        ("distro", {"name": "testdistroadd", "kernel": "Set in method", "initrd": "Set in method", "breed": "suse",
                    "arch": "x86_64"}),
        ("profile", {"name": "testprofileadd", "distro": "testdistroadd_profile"}),
        ("system", {"name": "testsystemadd", "profile": "testprofileadd_system"}),
        ("image", {"name": "testimageadd"}),
        ("repo", {"name": "testrepoadd", "mirror": "http://localhost"}),
        ("package", {"name": "testpackageadd"}),
        ("mgmtclass", {"name": "testmgmtclassadd"}),
        ("file", {"name": "testfileadd", "path": "/tmp", "owner": "root", "group": "root", "mode": "600",
                  "is-dir": "True"}),
        ("menu", {"name": "testmenuadd"}),
    ])
    def test_add(self, run_cmd, remove_object_via_cli, generate_run_cmd_array, create_kernel_initrd, fk_initrd,
                 fk_kernel, add_object_via_cli, object_type, attributes):
        # Arrange
        folder = create_kernel_initrd(fk_kernel, fk_initrd)
        kernel_path = os.path.join(folder, fk_kernel)
        initrd_path = os.path.join(folder, fk_kernel)
        name_distro_profile = "testdistroadd_profile"
        name_distro_system = "testdistroadd_system"
        name_profile_system = "testprofileadd_system"
        if object_type == "distro":
            attributes["kernel"] = kernel_path
            attributes["initrd"] = initrd_path
        elif object_type == "profile":
            add_object_via_cli("distro", {"name": name_distro_profile, "kernel": kernel_path, "initrd": initrd_path,
                                          "breed": "suse", "arch": "x86_64"})
        elif object_type == "system":
            add_object_via_cli("distro", {"name": name_distro_system, "kernel": kernel_path, "initrd": initrd_path,
                                          "breed": "suse", "arch": "x86_64"})
            add_object_via_cli("profile", {"name": name_profile_system, "distro": name_distro_system})

        cmd_list = [object_type, "add"]
        options = generate_run_cmd_array(attributes)
        cmd_list.extend(options)

        # Act
        (outputstd, outputerr) = run_cmd(cmd=cmd_list)

        # Cleanup
        remove_object_via_cli(object_type, attributes["name"])
        if object_type == "profile":
            remove_object_via_cli("distro", name_distro_profile)
        elif object_type == "system":
            remove_object_via_cli("profile", name_profile_system)
            remove_object_via_cli("distro", name_distro_system)

        # Assert
        assert not outputstd

    @pytest.mark.parametrize("object_type,attributes", [
        ("distro", {"name": "testdistroremove", "kernel": "Set in method", "initrd": "Set in method", "breed": "suse",
                    "arch": "x86_64"}),
        ("profile", {"name": "testprofileremove", "distro": "testdistroremove_profile"}),
        ("system", {"name": "testsystemremove", "profile": "testprofileremove_system"}),
        ("image", {"name": "testimageremove"}),
        ("repo", {"name": "testreporemove", "mirror": "http://localhost"}),
        ("package", {"name": "testpackageremove"}),
        ("mgmtclass", {"name": "testmgmtclassremove"}),
        ("file", {"name": "testfileremove", "path": "/tmp", "owner": "root", "group": "root", "mode": "600",
                  "is-dir": "True"}),
        ("menu", {"name": "testmenuremove"}),
    ])
    def test_remove(self, run_cmd, add_object_via_cli, remove_object_via_cli, create_kernel_initrd, fk_initrd,
                    fk_kernel, object_type, attributes):
        # Arrange
        folder = create_kernel_initrd(fk_kernel, fk_initrd)
        kernel_path = os.path.join(folder, fk_kernel)
        initrd_path = os.path.join(folder, fk_kernel)
        name_distro_profile = "testdistroremove_profile"
        name_distro_system = "testdistroremove_system"
        name_profile_system = "testprofileremove_system"
        if object_type == "distro":
            attributes["kernel"] = kernel_path
            attributes["initrd"] = initrd_path
        elif object_type == "profile":
            add_object_via_cli("distro", {"name": name_distro_profile, "kernel": kernel_path, "initrd": initrd_path,
                                          "breed": "suse", "arch": "x86_64"})
        elif object_type == "system":
            add_object_via_cli("distro", {"name": name_distro_system, "kernel": kernel_path, "initrd": initrd_path,
                                          "breed": "suse", "arch": "x86_64"})
            add_object_via_cli("profile", {"name": name_profile_system, "distro": name_distro_system})
        add_object_via_cli(object_type, attributes)

        # Act
        (outputstd, outputerr) = run_cmd(cmd=[object_type, "remove", "--name=%s" % attributes["name"]])

        # Cleanup
        if object_type == "profile":
            remove_object_via_cli("distro", name_distro_profile)
        elif object_type == "system":
            remove_object_via_cli("profile", name_profile_system)
            remove_object_via_cli("distro", name_distro_system)

        # Assert
        assert not outputstd



"""
Reports on automatic installation activity by examining the logs in
/var/log/cobbler.

Copyright 2007-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""

import glob
import gzip
import re
import time

# ARRAY INDEXES
MOST_RECENT_START = 0
MOST_RECENT_STOP = 1
MOST_RECENT_TARGET = 2
SEEN_START = 3
SEEN_STOP = 4
STATE = 5


class CobblerStatusReport:

    def __init__(self, collection_mgr, mode: str):
        """
        Constructor

        :param collection_mgr: The collection manager which holds all information.
        :param mode: This describes how Cobbler should report. Currently there only the option ``text`` can be set
                     explicitly.
        """
        self.collection_mgr = collection_mgr
        self.settings = collection_mgr.settings()
        self.ip_data = {}
        self.mode = mode

    # -------------------------------------------------------

    def scan_logfiles(self):
        """
        Scan the install log-files - starting with the oldest file.
        """
        unsorted_files = glob.glob("/var/log/cobbler/install.log*")
        files_dict = dict()
        log_id_re = re.compile(r'install.log.(\d+)')
        for fname in unsorted_files:
            id_match = log_id_re.search(fname)
            if id_match:
                files_dict[int(id_match.group(1))] = fname

        files = list()
        sorted_ids = sorted(files_dict, key=files_dict.get, reverse=True)
        for file_id in sorted_ids:
            files.append(files_dict[file_id])
        if '/var/log/cobbler/install.log' in unsorted_files:
            files.append('/var/log/cobbler/install.log')

        for fname in files:
            if fname.endswith('.gz'):
                fd = gzip.open(fname)
            else:
                fd = open(fname)
            data = fd.read()
            for line in data.split("\n"):
                tokens = line.split()
                if len(tokens) == 0:
                    continue
                (profile_or_system, name, ip, start_or_stop, ts) = tokens
                self.catalog(profile_or_system, name, ip, start_or_stop, ts)
            fd.close()

    # ------------------------------------------------------

    def catalog(self, profile_or_system: str, name: str, ip, start_or_stop: str, ts: float):
        """
        Add a system to ``cobbler status``.

        :param profile_or_system: This can be ``system`` or ``profile``.
        :param name: The name of the object.
        :param ip: The ip of the system to watch.
        :param start_or_stop: This parameter may be ``start`` or ``stop``
        :param ts: Don't know what this does.
        """
        ip_data = self.ip_data

        if ip not in ip_data:
            ip_data[ip] = [-1, -1, "?", 0, 0, "?"]
        elem = ip_data[ip]

        ts = float(ts)

        mrstart = elem[MOST_RECENT_START]
        mrstop = elem[MOST_RECENT_STOP]
        mrtarg = elem[MOST_RECENT_TARGET]

        if start_or_stop == "start":
            if mrstart < ts:
                mrstart = ts
                mrtarg = "%s:%s" % (profile_or_system, name)
                elem[SEEN_START] += 1

        if start_or_stop == "stop":
            if mrstop < ts:
                mrstop = ts
                mrtarg = "%s:%s" % (profile_or_system, name)
                elem[SEEN_STOP] += 1

        elem[MOST_RECENT_START] = mrstart
        elem[MOST_RECENT_STOP] = mrstop
        elem[MOST_RECENT_TARGET] = mrtarg

    # -------------------------------------------------------

    def process_results(self):
        """
        Look through all systems which were collected and update the status.

        :return: Return ``ip_data`` of the object.
        """
        # FIXME: this should update the times here
        tnow = int(time.time())
        for ip in list(self.ip_data.keys()):
            elem = self.ip_data[ip]
            start = int(elem[MOST_RECENT_START])
            stop = int(elem[MOST_RECENT_STOP])
            if stop > start:
                elem[STATE] = "finished"
            else:
                delta = tnow - start
                min = delta // 60
                sec = delta % 60
                if min > 100:
                    elem[STATE] = "unknown/stalled"
                else:
                    elem[STATE] = "installing (%sm %ss)" % (min, sec)

        return self.ip_data

    def get_printable_results(self):
        """
        Convert the status of Cobbler from a machine readable form to human readable.

        :return: A nice formatted representation of the results of ``cobbler status``.
        """
        format = "%-15s|%-20s|%-17s|%-17s"
        ip_data = self.ip_data
        ips = list(ip_data.keys())
        ips.sort()
        line = (
            "ip",
            "target",
            "start",
            "state",
        )
        buf = format % line
        for ip in ips:
            elem = ip_data[ip]
            if elem[MOST_RECENT_START] > -1:
                start = time.ctime(elem[MOST_RECENT_START])
            else:
                start = "Unknown"
            line = (
                ip,
                elem[MOST_RECENT_TARGET],
                start,
                elem[STATE]
            )
            buf += "\n" + format % line
        return buf

    # -------------------------------------------------------

    def run(self):
        """
        Calculate and print a automatic installation status report.
        """
        self.scan_logfiles()
        results = self.process_results()
        if self.mode == "text":
            return self.get_printable_results()
        else:
            return results



"""
Builds out filesystem trees/data based on the object tree.
This is the code behind 'cobbler sync'.

Copyright 2006-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""

import glob
import logging
import os
import time
from typing import Optional, List

from cobbler.cexceptions import CX
from cobbler import templar
from cobbler import tftpgen
from cobbler import utils


class CobblerSync:
    """
    Handles conversion of internal state to the tftpboot tree layout
    """

    def __init__(self, collection_mgr, verbose: bool = True, dhcp=None, dns=None, tftpd=None):
        """
        Constructor

        :param collection_mgr: The collection manager instance which holds all information about cobbler.
        :param verbose: Whether to log the actions performed in this module verbose or not.
        :param dhcp: The DHCP manager which can update the DHCP config.
        :param dns: The DNS manager which can update the DNS config.
        :param tftpd: The TFTP manager which can update the TFTP config.
        """
        self.logger = logging.getLogger()

        self.verbose = verbose
        self.collection_mgr = collection_mgr
        self.api = collection_mgr.api
        self.distros = collection_mgr.distros()
        self.profiles = collection_mgr.profiles()
        self.systems = collection_mgr.systems()
        self.images = collection_mgr.images()
        self.settings = collection_mgr.settings()
        self.repos = collection_mgr.repos()
        self.templar = templar.Templar(collection_mgr)
        self.tftpgen = tftpgen.TFTPGen(collection_mgr)
        self.dns = dns
        self.dhcp = dhcp
        self.tftpd = tftpd
        self.bootloc = self.settings.tftpboot_location

        self.pxelinux_dir = os.path.join(self.bootloc, "pxelinux.cfg")
        self.grub_dir = os.path.join(self.bootloc, "grub")
        self.images_dir = os.path.join(self.bootloc, "images")
        self.ipxe_dir = os.path.join(self.bootloc, "ipxe")
        self.rendered_dir = os.path.join(self.settings.webdir, "rendered")
        self.links = os.path.join(self.settings.webdir, "links")
        self.distromirror_config = os.path.join(self.settings.webdir, "distro_mirror/config")
        # FIXME: See https://github.com/cobbler/cobbler/issues/2453
        # Move __create_tftpboot_dirs() outside of sync.py.
        self.__create_tftpboot_dirs()

    def run_sync_systems(self, systems: List[str]):
        """
        Syncs the specific systems with the config tree.
        """
        if not os.path.exists(self.bootloc):
            utils.die("cannot find directory: %s" % self.bootloc)

        self.logger.info("running pre-sync triggers")

        # run pre-triggers...
        utils.run_triggers(self.api, None, "/var/lib/cobbler/triggers/sync/pre/*")

        self.distros = self.collection_mgr.distros()
        self.profiles = self.collection_mgr.profiles()
        self.systems = self.collection_mgr.systems()
        self.settings = self.collection_mgr.settings()
        self.repos = self.collection_mgr.repos()

        # Have the tftpd module handle copying bootloaders, distros, images, and all_system_files
        self.tftpd.sync_systems(systems)

        if self.settings.manage_dhcp:
            self.write_dhcp()
        if self.settings.manage_dns:
            self.logger.info("rendering DNS files")
            self.dns.regen_hosts()
            self.dns.write_dns_files()

        self.logger.info("cleaning link caches")
        self.clean_link_cache()

        if self.settings.manage_rsync:
            self.logger.info("rendering rsync files")
            self.rsync_gen()

        # run post-triggers
        self.logger.info("running post-sync triggers")
        utils.run_triggers(self.api, None, "/var/lib/cobbler/triggers/sync/post/*")
        utils.run_triggers(self.api, None, "/var/lib/cobbler/triggers/change/*")

    def run(self):
        """
        Syncs the current configuration file with the config tree.
        Using the ``Check().run_`` functions previously is recommended
        """
        if not os.path.exists(self.bootloc):
            utils.die("cannot find directory: %s" % self.bootloc)

        self.logger.info("running pre-sync triggers")

        # run pre-triggers...
        utils.run_triggers(self.api, None, "/var/lib/cobbler/triggers/sync/pre/*")

        self.distros = self.collection_mgr.distros()
        self.profiles = self.collection_mgr.profiles()
        self.systems = self.collection_mgr.systems()
        self.settings = self.collection_mgr.settings()
        self.repos = self.collection_mgr.repos()

        # execute the core of the sync operation
        self.logger.info("cleaning trees")
        self.clean_trees()

        # Have the tftpd module handle copying bootloaders, distros, images, and all_system_files
        self.tftpd.sync(self.verbose)
        # Copy distros to the webdir
        # Adding in the exception handling to not blow up if files have been moved (or the path references an NFS
        # directory that's no longer mounted)
        for d in self.distros:
            try:
                self.logger.info("copying files for distro: %s" % d.name)
                self.tftpgen.copy_single_distro_files(d, self.settings.webdir, True)
                self.tftpgen.write_templates(d, write_file=True)
            except CX as e:
                self.logger.error(e.value)

        # make the default pxe menu anyway...
        self.tftpgen.make_pxe_menu()

        if self.settings.manage_dhcp:
            self.write_dhcp()
        if self.settings.manage_dns:
            self.logger.info("rendering DNS files")
            self.dns.regen_hosts()
            self.dns.write_configs()

        if self.settings.manage_tftpd:
            # copy in boot_files
            self.tftpd.write_boot_files()

        self.logger.info("cleaning link caches")
        self.clean_link_cache()

        if self.settings.manage_rsync:
            self.logger.info("rendering Rsync files")
            self.rsync_gen()

        # run post-triggers
        self.logger.info("running post-sync triggers")
        utils.run_triggers(self.api, None, "/var/lib/cobbler/triggers/sync/post/*")
        utils.run_triggers(self.api, None, "/var/lib/cobbler/triggers/change/*")

    def __create_tftpboot_dirs(self):
        """
        Create directories for tftpboot images
        """
        if not os.path.exists(self.pxelinux_dir):
            utils.mkdir(self.pxelinux_dir)
        if not os.path.exists(self.grub_dir):
            utils.mkdir(self.grub_dir)
        grub_images_link = os.path.join(self.grub_dir, "images")
        if not os.path.exists(grub_images_link):
            os.symlink("../images", grub_images_link)
        if not os.path.exists(self.images_dir):
            utils.mkdir(self.images_dir)
        if not os.path.exists(self.rendered_dir):
            utils.mkdir(self.rendered_dir)
        if not os.path.exists(self.ipxe_dir):
            utils.mkdir(self.ipxe_dir)
        if not os.path.exists(self.links):
            utils.mkdir(self.links)
        if not os.path.exists(self.distromirror_config):
            utils.mkdir(self.distromirror_config)

    def clean_trees(self):
        """
        Delete any previously built pxelinux.cfg tree and virt tree info and then create directories.

        Note: for SELinux reasons, some information goes in ``/tftpboot``, some in ``/var/www/cobbler`` and some must be
        duplicated in both. This is because PXE needs tftp, and automatic installation and Virt operations need http.
        Only the kernel and initrd images are duplicated, which is unfortunate, though SELinux won't let me give them
        two contexts, so symlinks are not a solution. *Otherwise* duplication is minimal.
        """

        # clean out parts of webdir and all of /tftpboot/images and /tftpboot/pxelinux.cfg
        for x in os.listdir(self.settings.webdir):
            path = os.path.join(self.settings.webdir, x)
            if os.path.isfile(path):
                if not x.endswith(".py"):
                    utils.rmfile(path)
            if os.path.isdir(path):
                if x not in self.settings.webdir_whitelist:
                    # delete directories that shouldn't exist
                    utils.rmtree(path)
                if x in ["templates", "images", "systems", "distros", "profiles", "repo_profile", "repo_system",
                         "rendered"]:
                    # clean out directory contents
                    utils.rmtree_contents(path)
        self.__create_tftpboot_dirs()
        utils.rmtree_contents(self.pxelinux_dir)
        utils.rmtree_contents(self.grub_dir)
        utils.rmtree_contents(self.images_dir)
        utils.rmtree_contents(self.ipxe_dir)
        utils.rmtree_contents(self.rendered_dir)

    def write_dhcp(self):
        """
        Write all files which are associated to DHCP.
        """
        self.logger.info("rendering DHCP files")
        self.dhcp.write_configs()
        self.dhcp.regen_ethers()

    def sync_dhcp(self):
        """
        This calls write_dhcp and restarts the DHCP server.
        """
        if self.settings.manage_dhcp:
            self.write_dhcp()
            self.dhcp.sync_dhcp()

    def clean_link_cache(self):
        """
        All files which are linked into the cache will be deleted so the cache can be rebuild.
        """
        for dirtree in [os.path.join(self.bootloc, 'images'), self.settings.webdir]:
            cachedir = os.path.join(dirtree, '.link_cache')
            if os.path.isdir(cachedir):
                cmd = "find %s -maxdepth 1 -type f -links 1 -exec rm -f '{}' ';'" % cachedir
                utils.subprocess_call(cmd)

    def rsync_gen(self):
        """
        Generate rsync modules of all repositories and distributions

        :raises OSError:
        """
        template_file = "/etc/cobbler/rsync.template"

        try:
            template = open(template_file, "r")
        except:
            raise OSError("error reading template %s" % template_file)

        template_data = ""
        template_data = template.read()
        template.close()

        distros = []

        for link in glob.glob(os.path.join(self.settings.webdir, 'links', '*')):
            distro = {}
            distro["path"] = os.path.realpath(link)
            distro["name"] = os.path.basename(link)
            distros.append(distro)

        repos = [repo.name for repo in self.api.repos()
                 if os.path.isdir(os.path.join(self.settings.webdir, "repo_mirror", repo.name))]

        metadata = {
            "date": time.asctime(time.gmtime()),
            "cobbler_server": self.settings.server,
            "distros": distros,
            "repos": repos,
            "webdir": self.settings.webdir
        }

        self.templar.render(template_data, metadata, "/etc/rsyncd.conf")

    def add_single_distro(self, name):
        """
        Sync adding a single distro.

        :param name: The name of the distribution.
        """
        # get the distro record
        distro = self.distros.find(name=name)
        if distro is None:
            return
        # copy image files to images/$name in webdir & tftpboot:
        self.tftpgen.copy_single_distro_files(distro, self.settings.webdir, True)
        self.tftpd.add_single_distro(distro)

        # create the symlink for this distro
        src_dir = utils.find_distro_path(self.settings, distro)
        dst_dir = os.path.join(self.settings.webdir, "links", name)
        if os.path.exists(dst_dir):
            self.logger.warning("skipping symlink, destination (%s) exists", dst_dir)
        elif utils.path_tail(os.path.join(self.settings.webdir, "distro_mirror"), src_dir) == "":
            self.logger.warning("skipping symlink, the source (%s) is not in %s",
                                src_dir, os.path.join(self.settings.webdir, "distro_mirror"))
        else:
            try:
                self.logger.info("trying symlink %s -> %s", src_dir, dst_dir)
                os.symlink(src_dir, dst_dir)
            except (IOError, OSError):
                self.logger.error("symlink failed (%s -> %s)", src_dir, dst_dir)

        # generate any templates listed in the distro
        self.tftpgen.write_templates(distro, write_file=True)
        # cascade sync
        kids = distro.get_children()
        for k in kids:
            self.add_single_profile(k, rebuild_menu=False)
        self.tftpgen.make_pxe_menu()

    def add_single_image(self, name):
        """
        Sync adding a single image.

        :param name: The name of the image.
        """
        image = self.images.find(name=name)
        self.tftpgen.copy_single_image_files(image)
        kids = image.get_children()
        for k in kids:
            self.add_single_system(k)
        self.tftpgen.make_pxe_menu()

    def remove_single_distro(self, name):
        """
        Sync removing a single distro.

        :param name: The name of the distribution.
        """
        bootloc = self.settings.tftpboot_location
        # delete contents of images/$name directory in webdir
        utils.rmtree(os.path.join(self.settings.webdir, "images", name))
        # delete contents of images/$name in tftpboot
        utils.rmtree(os.path.join(bootloc, "images", name))
        # delete potential symlink to tree in webdir/links
        utils.rmfile(os.path.join(self.settings.webdir, "links", name))
        # delete potential distro config files
        utils.rmglob_files(os.path.join(self.settings.webdir, "distro_mirror", "config"), name + "*.repo")

    def remove_single_image(self, name):
        """
        Sync removing a single image.

        :param name: The name of the image.
        """
        bootloc = self.settings.tftpboot_location
        utils.rmfile(os.path.join(bootloc, "images2", name))

    def add_single_profile(self, name: str, rebuild_menu: bool = True) -> Optional[bool]:
        """
        Sync adding a single profile.

        :param name: The name of the profile.
        :param rebuild_menu: Whether to rebuild the grub/... menu or not.
        :return: ``True`` if this succeeded.
        """
        # get the profile object:
        profile = self.profiles.find(name=name)
        if profile is None:
            # Most likely a subprofile's kid has been removed already, though the object tree has not been reloaded and
            # this is just noise.
            return
        # Rebuild the yum configuration files for any attached repos generate any templates listed in the distro.
        self.tftpgen.write_templates(profile)
        # Cascade sync
        kids = profile.children
        for k in kids:
            if self.api.find_profile(name=k) is not None:
                self.add_single_profile(k, rebuild_menu=False)
            else:
                self.add_single_system(k)
        if rebuild_menu:
            self.tftpgen.make_pxe_menu()
        return True

    def remove_single_profile(self, name: str, rebuild_menu: bool = True):
        """
        Sync removing a single profile.

        :param name: The name of the profile.
        :param rebuild_menu: Whether to rebuild the grub/... menu or not.
        """
        # delete profiles/$name file in webdir
        utils.rmfile(os.path.join(self.settings.webdir, "profiles", name))
        # delete contents on autoinstalls/$name directory in webdir
        utils.rmtree(os.path.join(self.settings.webdir, "autoinstalls", name))
        if rebuild_menu:
            self.tftpgen.make_pxe_menu()

    def update_system_netboot_status(self, name: str):
        """
        Update the netboot status of a system.

        :param name: The name of the system.
        """
        system = self.systems.find(name=name)
        if system is None:
            return
        self.tftpd.sync_single_system(system)

    def add_single_system(self, name: str):
        """
        Sync adding a single system.

        :param name: The name of the system.
        """
        # get the system object:
        system = self.systems.find(name=name)
        if system is None:
            return
        # rebuild system_list file in webdir
        if self.settings.manage_dhcp:
            self.dhcp.regen_ethers()
        if self.settings.manage_dns:
            self.dns.regen_hosts()
        # write the PXE files for the system
        self.tftpd.sync_single_system(system)

    def remove_single_system(self, name: str):
        """
        Sync removing a single system.

        :param name: The name of the system.
        """
        bootloc = self.settings.tftpboot_location
        # delete contents of autoinsts_sys/$name in webdir
        system_record = self.systems.find(name=name)

        for (name, interface) in list(system_record.interfaces.items()):
            pxe_filename = system_record.get_config_filename(interface=name, loader="pxe")
            grub_filename = system_record.get_config_filename(interface=name, loader="grub")
            utils.rmfile(os.path.join(bootloc, "pxelinux.cfg", pxe_filename))
            utils.rmfile(os.path.join(bootloc, "grub", "system", grub_filename))
            utils.rmfile(os.path.join(bootloc, "grub", "system_link", system_record.name))

    def remove_single_menu(self, rebuild_menu: bool = True):
        """
        Sync removing a single menu.
        :param rebuild_menu: Whether to rebuild the grub/... menu or not.
        """
        if rebuild_menu:
            self.tftpgen.make_pxe_menu()



"""
The action module is responsible for containing one Python module for each action which Cobbler offers. The code should
never be dependent on another module or on other parts. An action should request the exact
data it requires and nothing more.
"""



"""
Custom exceptions for Cobbler

Copyright 2006-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""


class CobblerException(Exception):
    """
    This is the default Cobbler exception where all other exceptions are inheriting from.
    """

    def __init__(self, value, *args):
        """
        Default constructor for the Exception.

        Bad example: ``CobblerException("Profile %s not found" % profile_name)``

        Good example: ``CobblerException("Profile %s not found", profile_name)``

        :param value: The string representation of the Exception. Do not glue strings and pass them as one. Instead pass
                      them as params and let the constructor of the Exception build the string like (same as it should
                      be done with logging calls). Example see above.
        :param args: Optional arguments which replace a ``%s`` in a Python string.
        """
        self.value = value % args
        # this is a hack to work around some odd exception handling in older pythons
        self.from_cobbler = 1

    def __str__(self):
        """
        This is the string representation of the base Cobbler Exception.
        :return: self.value as a string represented.
        """
        return repr(self.value)


class CX(CobblerException):
    """
    This is a general exception which gets thrown often inside Cobbler.
    """
    pass



"""
Cobbler's file-based object serializer.
As of 9/2014, this is Cobbler's default serializer and the most stable one.
It uses multiple JSON files in /var/lib/cobbler/collections/distros, profiles, etc

Copyright 2006-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""

import os
import glob
import json

import cobbler.api as capi
from cobbler import settings
from cobbler.cexceptions import CX

libpath = "/var/lib/cobbler/collections"


def register() -> str:
    """
    The mandatory Cobbler module registration hook.
    """
    return "serializer"


def what() -> str:
    """
    Module identification function
    """
    return "serializer/file"


def __find_double_json_files(filename: str):
    """
    Finds a file with duplicate .json ending and renames it.
    :param filename: Filename to be checked
    :raises FileExistsError: If both JSON files exist
    """

    if not os.path.isfile(filename):
        if os.path.isfile(filename + ".json"):
            os.rename(filename + ".json", filename)
    else:
        if os.path.isfile(filename + ".json"):
            raise FileExistsError("Both JSON files (%s) exist!" % filename)


def serialize_item(collection, item):
    """
    Save a collection item to file system

    :param collection: collection
    :param item: collection item
    """

    if not item.name:
        raise CX("name unset for item!")

    collection_types = collection.collection_types()
    filename = os.path.join(libpath, collection_types, item.name + ".json")
    __find_double_json_files(filename)

    if capi.CobblerAPI().settings().serializer_pretty_json:
        sort_keys = True
        indent = 4
    else:
        sort_keys = False
        indent = None

    _dict = item.serialize()
    with open(filename, "w+") as file_descriptor:
        data = json.dumps(_dict, sort_keys=sort_keys, indent=indent)
        file_descriptor.write(data)


def serialize_delete(collection, item):
    """
    Delete a collection item from file system.

    :param collection: collection
    :param item: collection item
    """

    collection_types = collection.collection_types()
    filename = os.path.join(libpath, collection_types, item.name + ".json")
    __find_double_json_files(filename)

    if os.path.exists(filename):
        os.remove(filename)


def serialize(collection):
    """
    Save a collection to file system

    :param collection: collection
    """

    # do not serialize settings
    if collection.collection_type() != "setting":
        for x in collection:
            serialize_item(collection, x)


def deserialize_raw(collection_types: str):
    """
    Loads a collection from the disk.

    :param collection_types: The type of collection to load.
    :return: The loaded dictionary.
    """
    if collection_types == "settings":
        return settings.read_settings_file()
    else:
        results = []

        path = os.path.join(libpath, collection_types)
        all_files = glob.glob("%s/*.json" % path)

        for f in all_files:
            with open(f) as file_descriptor:
                json_data = file_descriptor.read()
                _dict = json.loads(json_data)
                results.append(_dict)
        return results


def deserialize(collection, topological: bool = True):
    """
    Load a collection from file system.

    :param collection: The collection to deserialize.
    :param topological: If the collection list should be sorted by the
                        collection dict key 'depth' value or not.
    """

    datastruct = deserialize_raw(collection.collection_types())
    if topological and isinstance(datastruct, list):
        datastruct.sort(key=lambda x: x.get("depth", 1))
    if isinstance(datastruct, dict):
        collection.from_dict(datastruct)
    elif isinstance(datastruct, list):
        collection.from_list(datastruct)



"""
Mod Python service functions for Cobbler's public interface
(aka cool stuff that works with wget/curl)

based on code copyright 2007 Albert P. Tobey <tobert@gmail.com>
additions: 2007-2009 Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""

import json
import time
import xmlrpc.client
import yaml

from cobbler import download_manager
from cobbler.cobbler_collections import manager
from cobbler.settings import Settings


class CobblerSvc:
    """
    Interesting mod python functions are all keyed off the parameter mode, which defaults to index. All options are
    passed as parameters into the function.
    """
    def __init__(self, server=None, req=None):
        """
        Default constructor which sets up everything to be ready.

        :param server: The domain to run at.
        :param req: This parameter is unused.
        """
        # ToDo: Remove req attribute.
        self.server = server
        self.remote = None
        self.req = req
        self.collection_mgr = manager.CollectionManager(self)
        self.dlmgr = download_manager.DownloadManager(self.collection_mgr)

    def __xmlrpc_setup(self):
        """
        Sets up the connection to the Cobbler XMLRPC server. This is the version that does not require a login.
        """
        if self.remote is None:
            self.remote = xmlrpc.client.Server(self.server, allow_none=True)

    def settings(self) -> Settings:
        """
        Get the application configuration.

        :return: Settings object.
        """
        self.__xmlrpc_setup()
        return Settings().from_dict(self.remote.get_settings())

    def index(self, **args) -> str:
        """
        Just a placeholder method as an entry point.

        :param args: This parameter is unused.
        :return: "no mode specified"
        """
        return "no mode specified"

    def debug(self, profile=None, **rest):
        # The purpose of this method could change at any time and is intented for temporary test code only, don't rely
        # on it.
        self.__xmlrpc_setup()
        return self.remote.get_repos_compatible_with_profile(profile)

    def autoinstall(self, profile=None, system=None, REMOTE_ADDR=None, REMOTE_MAC=None, **rest):
        """
        Generate automatic installation files.

        :param profile:
        :param system:
        :param REMOTE_ADDR:
        :param REMOTE_MAC:
        :param rest: This parameter is unused.
        :return:
        """
        self.__xmlrpc_setup()
        data = self.remote.generate_autoinstall(profile, system, REMOTE_ADDR, REMOTE_MAC)
        return "%s" % data

    def ks(self, profile=None, system=None, REMOTE_ADDR=None, REMOTE_MAC=None, **rest):
        """
        Generate automatic installation files. This is a legacy function for part backward compability to 2.6.6
        releases.

        :param profile:
        :param system:
        :param REMOTE_ADDR:
        :param REMOTE_MAC:
        :param rest: This parameter is unused.
        :return:
        """
        self.__xmlrpc_setup()
        data = self.remote.generate_autoinstall(profile, system, REMOTE_ADDR, REMOTE_MAC)
        return "%s" % data

    def ipxe(self, profile=None, image=None, system=None, mac=None, **rest):
        """
        Generates an iPXE configuration.

        :param profile: A profile.
        :param image: An image.
        :param system: A system.
        :param mac: A MAC address.
        :param rest: This parameter is unused.
        """
        self.__xmlrpc_setup()
        if not system and mac:
            query = {"mac_address": mac}
            if profile:
                query["profile"] = profile
            elif image:
                query["image"] = image
            found = self.remote.find_system(query)
            if found:
                system = found[0]

        data = self.remote.generate_ipxe(profile, image, system)
        return "%s" % data

    def bootcfg(self, profile=None, system=None, **rest):
        """
        Generate a boot.cfg config file. Used primarily for VMware ESXi.

        :param profile:
        :param system:
        :param rest: This parameter is unused.
        :return:
        """
        self.__xmlrpc_setup()
        data = self.remote.generate_bootcfg(profile, system)
        return "%s" % data

    def script(self, profile=None, system=None, **rest) -> str:
        """
        Generate a script based on snippets. Useful for post or late-action scripts where it's difficult to embed the
        script in the response file.

        :param profile: The profile to generate the script for.
        :param system: The system to generate the script for.
        :param rest: This may contain a parameter with the key "query_string" which has a key "script" which may be an
                     array. The element from position zero is taken.
        :return: The generated script.
        """
        self.__xmlrpc_setup()
        data = self.remote.generate_script(profile, system, rest['query_string']['script'][0])
        return "%s" % data

    def events(self, user="", **rest) -> str:
        """
        If no user is given then all events are returned. Otherwise only event associated to a user are returned.

        :param user: Filter the events for a given user.
        :param rest: This parameter is unused.
        :return: A JSON object which contains all events.
        """
        self.__xmlrpc_setup()
        if user == "":
            data = self.remote.get_events("")
        else:
            data = self.remote.get_events(user)

        # sort it... it looks like { timestamp : [ array of details ] }
        keylist = list(data.keys())
        keylist.sort()
        results = []
        for k in keylist:
            etime = int(data[k][0])
            nowtime = time.time()
            if ((nowtime - etime) < 30):
                results.append([k, data[k][0], data[k][1], data[k][2]])
        return json.dumps(results)

    def template(self, profile=None, system=None, path=None, **rest) -> str:
        """
        Generate a templated file for the system. Either specify a profile OR a system.

        :param profile: The profile to provide for the generation of the template.
        :param system: The system to provide for the generation of the template.
        :param path: The path to the template.
        :param rest: This parameter is unused.
        :return: The rendered template.
        """
        self.__xmlrpc_setup()
        if path is not None:
            path = path.replace("_", "/")
            path = path.replace("//", "_")
        else:
            return "# must specify a template path"

        if profile is not None:
            data = self.remote.get_template_file_for_profile(profile, path)
        elif system is not None:
            data = self.remote.get_template_file_for_system(system, path)
        else:
            data = "# must specify profile or system name"
        return data

    def yum(self, profile=None, system=None, **rest) -> str:
        """
        Generate a repo config. Either specify a profile OR a system.

        :param profile: The profile to provide for the generation of the template.
        :param system: The system to provide for the generation of the template.
        :param rest: This parameter is unused.
        :return: The generated repository config.
        """
        self.__xmlrpc_setup()
        if profile is not None:
            data = self.remote.get_repo_config_for_profile(profile)
        elif system is not None:
            data = self.remote.get_repo_config_for_system(system)
        else:
            data = "# must specify profile or system name"
        return data

    def trig(self, mode: str = "?", profile=None, system=None, REMOTE_ADDR=None, **rest) -> str:
        """
        Hook to call install triggers. Only valid for a profile OR a system.

        :param mode: Can be "pre", "post" or "firstboot". Everything else is invalid.
        :param profile: The profile object to run triggers for.
        :param system: The system object to run triggers for.
        :param REMOTE_ADDR: The ip if the remote system/profile.
        :param rest: This parameter is unused.
        :return: The return code of the action.
        """
        self.__xmlrpc_setup()
        ip = REMOTE_ADDR
        if profile:
            rc = self.remote.run_install_triggers(mode, "profile", profile, ip)
        else:
            rc = self.remote.run_install_triggers(mode, "system", system, ip)
        return str(rc)

    def nopxe(self, system=None, **rest) -> str:
        """
        Disables the network boot for the given system.

        :param system: The system to disable netboot for.
        :param rest: This parameter is unused.
        :return: A boolean status if the action succeed or not.
        """
        self.__xmlrpc_setup()
        return str(self.remote.disable_netboot(system))

    def list(self, what="systems", **rest) -> str:
        """
        Return a list of objects of a desired category. Defaults to "systems".

        :param what: May be "systems", "profiles", "distros", "images", "repos", "mgmtclasses", "packages",
                            "files" or "menus"
        :param rest: This parameter is unused.
        :return: The list of object names.
        """
        self.__xmlrpc_setup()
        buf = ""
        if what == "systems":
            listing = self.remote.get_systems()
        elif what == "profiles":
            listing = self.remote.get_profiles()
        elif what == "distros":
            listing = self.remote.get_distros()
        elif what == "images":
            listing = self.remote.get_images()
        elif what == "repos":
            listing = self.remote.get_repos()
        elif what == "mgmtclasses":
            listing = self.remote.get_mgmtclasses()
        elif what == "packages":
            listing = self.remote.get_packages()
        elif what == "files":
            listing = self.remote.get_files()
        elif what == "menus":
            listing = self.remote.get_menus()
        else:
            return "?"
        for x in listing:
            buf += "%s\n" % x["name"]
        return buf

    def autodetect(self, **rest) -> str:
        """
        This tries to autodect the system with the given information. If more than one candidate is found an error
        message is returned.

        :param rest: The keys "REMOTE_MACS", "REMOTE_ADDR" or "interfaces".
        :return: The name of the possible object or an error message.
        """
        self.__xmlrpc_setup()
        systems = self.remote.get_systems()

        # If kssendmac was in the kernel options line, see if a system can be found matching the MAC address. This is
        # more specific than an IP match.

        macinput = [mac.split(' ').lower() for mac in rest["REMOTE_MACS"]]

        ip = rest["REMOTE_ADDR"]

        candidates = []

        for x in systems:
            for y in x["interfaces"]:
                if x["interfaces"][y]["mac_address"].lower() in macinput:
                    candidates.append(x)

        if len(candidates) == 0:
            for x in systems:
                for y in x["interfaces"]:
                    if x["interfaces"][y]["ip_address"] == ip:
                        candidates.append(x)

        if len(candidates) == 0:
            return "FAILED: no match (%s,%s)" % (ip, macinput)
        elif len(candidates) > 1:
            return "FAILED: multiple matches"
        elif len(candidates) == 1:
            return candidates[0]["name"]

    def look(self, **rest) -> str:
        """
        Debug only: Show the handed dict via repr to the requester.
        :param rest: The dict to represent.
        :return: The dict reformated with repr()
        """

        return repr(rest)

    def find_autoinstall(self, system=None, profile=None, **rest):
        """
        Find an autoinstallation for a system or a profile. If this is not known different parameters can be passed to
        rest to find it automatically. See "autodetect".

        :param system: The system to find the autoinstallation for,
        :param profile: The profile to find the autoinstallation for.
        :param rest: The metadata to find the autoinstallation automatically.
        :return: The autoinstall script or error message.
        """
        self.__xmlrpc_setup()

        serverseg = "http://%s" % self.collection_mgr._settings.server

        name = "?"
        if system is not None:
            url = "%s/cblr/svc/op/autoinstall/system/%s" % (serverseg, name)
        elif profile is not None:
            url = "%s/cblr/svc/op/autoinstall/profile/%s" % (serverseg, name)
        else:
            name = self.autodetect(**rest)
            if name.startswith("FAILED"):
                return "# autodetection %s" % name
            url = "%s/cblr/svc/op/autoinstall/system/%s" % (serverseg, name)

        try:
            return self.dlmgr.urlread(url)
        except:
            return "# automatic installation file retrieval failed (%s)" % url

    def findks(self, system=None, profile=None, **rest):
        """
        This is a legacy function which enabled Cobbler partly to be backward compatible to 2.6.6 releases.

        It should be only be used if you must. Please use find_autoinstall if possible!
        :param system: If you wish to find a system please set this parameter to not null. Hand over the name of it.
        :param profile: If you wish to find a system please set this parameter to not null. Hand over the name of it.
        :param rest: If you wish you can try to let Cobbler autodetect the system with the MAC address.
        :return: Returns the autoinstall/kickstart profile.
        """
        self.__xmlrpc_setup()

        serverseg = "http://%s" % self.collection_mgr._settings.server

        name = "?"
        if system is not None:
            url = "%s/cblr/svc/op/ks/system/%s" % (serverseg, name)
        elif profile is not None:
            url = "%s/cblr/svc/op/ks/profile/%s" % (serverseg, name)
        else:
            name = self.autodetect(**rest)
            if name.startswith("FAILED"):
                return "# autodetection %s" % name
            url = "%s/cblr/svc/op/ks/system/%s" % (serverseg, name)

        try:
            return self.dlmgr.urlread(url)
        except:
            return "# kickstart retrieval failed (%s)" % url

    def puppet(self, hostname=None, **rest) -> str:
        """
        Dump the puppet data which is available for Cobbler.

        :param hostname: The hostname for the system which should the puppet data be dumped for.
        :param rest: This parameter is unused.
        :return: The yaml for the host.
        """
        self.__xmlrpc_setup()

        if hostname is None:
            return "hostname is required"

        settings = self.remote.get_settings()
        results = self.remote.find_system_by_dns_name(hostname)

        classes = results.get("mgmt_classes", {})
        params = results.get("mgmt_parameters", {})
        environ = results.get("status", "")

        data = {
            "classes": classes,
            "parameters": params,
            "environment": environ,
        }

        if environ == "":
            data.pop("environment", None)

        if settings.get("puppet_parameterized_classes", False):
            for ckey in list(classes.keys()):
                tmp = {}
                class_name = classes[ckey].get("class_name", "")
                if class_name in (None, ""):
                    class_name = ckey
                if classes[ckey].get("is_definition", False):
                    def_tmp = {}
                    def_name = classes[ckey]["params"].get("name", "")
                    del classes[ckey]["params"]["name"]
                    if def_name != "":
                        for pkey in list(classes[ckey]["params"].keys()):
                            def_tmp[pkey] = classes[ckey]["params"][pkey]
                        tmp["instances"] = {def_name: def_tmp}
                    else:
                        # FIXME: log an error here?
                        # skip silently...
                        continue
                else:
                    for pkey in list(classes[ckey]["params"].keys()):
                        tmp[pkey] = classes[ckey]["params"][pkey]
                del classes[ckey]
                classes[class_name] = tmp
        else:
            classes = list(classes.keys())

        return yaml.dump(data, default_flow_style=False)



"""
Cobbler provides builtin methods for use in Cheetah templates. $SNIPPET is one
such function and is now used to implement Cobbler's SNIPPET:: syntax.

Written by Daniel Guernsey <danpg102@gmail.com>
Contributions by Michael DeHaan <michael.dehaan AT gmail>
US Government work; No explicit copyright attached to this file.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""
import logging
import os.path
import re
from typing import Match, Optional, TextIO, Tuple, Union

from Cheetah.Template import Template

from cobbler import utils


# This class is defined using the Cheetah language. Using the 'compile' function we can compile the source directly into
# a Python class. This class will allow us to define the cheetah builtins.

logger = logging.getLogger()


def read_macro_file(location='/etc/cobbler/cheetah_macros'):
    if not os.path.exists(location):
        raise FileNotFoundError("Cobbler Cheetah Macros File must exist!")
    with open(location, "r") as macro_file:
        return macro_file.read()


def generate_cheetah_macros():
    try:
        macro_file = read_macro_file()
        return Template.compile(
            source=macro_file,
            moduleName="cobbler.template_api",
            className="CheetahMacros")
    except FileNotFoundError:
        logger.warning("Cheetah Macros file note found. Using empty template.")
        return Template.compile(source="")


class CobblerTemplate(generate_cheetah_macros()):
    """
    This class will allow us to include any pure python builtin functions.
    It derives from the cheetah-compiled class above. This way, we can include both types (cheetah and pure python) of
    builtins in the same base template. We don't need to override __init__
    """

    def __init__(self, **kwargs):
        """
        Constructor for this derived class. We include two additional default templates.

        :param kwargs: These arguments get passed to the super constructor of this class.
        """
        # This part (see 'Template' below for the other part) handles the actual inclusion of the file contents. We
        # still need to make the snippet's namespace (searchList) available to the template calling SNIPPET (done in
        # the other part).

        # This function can be used in two ways:
        # Cheetah syntax:
        # - $SNIPPET('my_snippet')
        # - SNIPPET syntax:
        # - SNIPPET::my_snippet

        # This follows all of the rules of snippets and advanced snippets. First it searches for a per-system snippet,
        # then a per-profile snippet, then a general snippet. If none is found, a comment explaining the error is
        # substituted.
        self.BuiltinTemplate = Template.compile(source="\n".join([
            "#def SNIPPET($file)",
            "#set $snippet = $read_snippet($file)",
            "#if $snippet",
            "#include source=$snippet",
            "#else",
            "# Error: no snippet data for $file",
            "#end if",
            "#end def",
        ]) + "\n")
        super().__init__(**kwargs)

    # OK, so this function gets called by Cheetah.Template.Template.__init__ to compile the template into a class. This
    # is probably a kludge, but it add a baseclass argument to the standard compile (see Cheetah's compile docstring)
    # and returns the resulting class. This argument, of course, points to this class. Now any methods entered here (or
    # in the base class above) will be accessible to all cheetah templates compiled by Cobbler.

    @classmethod
    def compile(cls, *args, **kwargs) -> bytes:
        """
        Compile a cheetah template with Cobbler modifications. Modifications include ``SNIPPET::`` syntax replacement
        and inclusion of Cobbler builtin methods. Please be aware that you cannot use the ``baseclass`` attribute of
        Cheetah anymore due to the fact that we are using it in our implementation to enable the Cheetah Macros.

        :param args: These just get passed right to Cheetah.
        :param kwargs: We just execute our own preprocessors and remove them and let afterwards handle Cheetah the rest.
        :return: The compiled template.
        """

        def replacer(match: Match):
            return "$SNIPPET('%s')" % match.group(1)

        def preprocess(source: Optional[str], file: Union[TextIO, str]) -> Tuple[str, Union[TextIO, str]]:
            # Normally, the cheetah compiler worries about this, but we need to preprocess the actual source.
            if source is None:
                if hasattr(file, 'read'):
                    source = file.read()
                else:
                    if os.path.exists(file):
                        with open(file, "r") as f:
                            source = "#errorCatcher Echo\n" + f.read()
                    else:
                        source = "# Unable to read %s\n" % file
                # Stop Cheetah from throwing a fit.
                file = None

            snippet_regex = re.compile(r'SNIPPET::([A-Za-z0-9_\-/.]+)')
            results = snippet_regex.sub(replacer, source)
            return results, file

        preprocessors = [preprocess]
        if 'preprocessors' in kwargs:
            preprocessors.extend(kwargs['preprocessors'])
        kwargs['preprocessors'] = preprocessors

        # Now let Cheetah do the actual compilation
        return super().compile(*args, **kwargs)

    def read_snippet(self, file: str) -> Optional[str]:
        """
        Locate the appropriate snippet for the current system and profile and read its contents.

        This file could be located in a remote location.

        This will first check for a per-system snippet, a per-profile snippet, a distro snippet, and a general snippet.

        :param file: The name of the file to read. Depending on the context this gets expanded automatically.
        :return: None (if the snippet file was not found) or the string with the read snippet.
        :raises AttributeError: Raised in case ``autoinstall_snippets_dir`` is missing.
        :raises FileNotFoundError: Raised in case some files are not found.
        """
        if not self.varExists('autoinstall_snippets_dir'):
            raise AttributeError("\"autoinstall_snippets_dir\" is required to find snippets")

        for snippet_class in ('system', 'profile', 'distro'):
            if self.varExists('%s_name' % snippet_class):
                full_path = '%s/per_%s/%s/%s' % (self.getVar('autoinstall_snippets_dir'), snippet_class, file,
                                                 self.getVar('%s_name' % snippet_class))
                try:
                    contents = utils.read_file_contents(full_path, fetch_if_remote=True)
                    return contents
                except FileNotFoundError:
                    pass

        try:
            full_path = '%s/%s' % (self.getVar('autoinstall_snippets_dir'), file)
            return "#errorCatcher ListErrors\n" + utils.read_file_contents(full_path, fetch_if_remote=True)
        except FileNotFoundError:
            return None

    def SNIPPET(self, file: str):
        """
        Include the contents of the named snippet here. This is equivalent to the #include directive in Cheetah, except
        that it searches for system and profile specific snippets, and it includes the snippet's namespace.

        This may be a little frobby, but it's really cool. This is a pure python portion of SNIPPET that appends the
        snippet's searchList to the caller's searchList. This makes any #defs within a given snippet available to the
        template that included the snippet.

        :param file: The snippet file to read and include in the template.
        :return: The updated template.
        """
        # First, do the actual inclusion. Cheetah (when processing #include) will track the inclusion in
        # self._CHEETAH__cheetahIncludes
        result = self.BuiltinTemplate.SNIPPET(self, file)

        # Now do our dirty work: locate the new include, and append its searchList to ours. We have to compute the full
        # path again? Eww.

        # This weird method is getting even weirder, the cheetah includes keys are no longer filenames but actual
        # contents of snippets. Regardless this seems to work and hopefully it will be ok.

        snippet_contents = self.read_snippet(file)
        if snippet_contents:
            # Only include what we don't already have. Because Cheetah passes our searchList into included templates,
            # the snippet's searchList will include this templates searchList. We need to avoid duplicating entries.
            childList = self._CHEETAH__cheetahIncludes[snippet_contents].searchList()
            myList = self.searchList()
            for childElem in childList:
                if childElem not in myList:
                    myList.append(childElem)

        return result

    # pylint: disable=R0201
    def sedesc(self, value: str) -> str:
        """
        Escape a string for use in sed.

        This function is used by several cheetah methods in cheetah_macros. It can be used by the end user as well.

        Example: Replace all instances of ``/etc/banner`` with a value stored in ``$new_banner``

        ..code::

           sed 's/$sedesc("/etc/banner")/$sedesc($new_banner)/'

        :param value: The phrase to escape.
        :return: The escaped phrase.
        """

        def escchar(character: str) -> str:
            if character in '/^.[]$()|*+?{}\\':
                return '\\' + character
            return character

        return ''.join([escchar(c) for c in value])



import pytest


@pytest.fixture
def create_menu(remote, token):
    """
    Creates a Menu "testmenu0" with a display_name "Test Menu0"

    :param remote: The xmlrpc object to connect to.
    :param token: The token to authenticate against the remote object.
    """
    menu = remote.new_menu(token)
    remote.modify_menu(menu, "name", "testmenu0", token)
    remote.modify_menu(menu, "display_name", "Test Menu0", token)
    remote.save_menu(menu, token)


@pytest.fixture
def remove_menu(remote, token):
    """
    Removes the Menu "testmenu0" which can be created with create_menu.

    :param remote: The xmlrpc object to connect to.
    :param token: The token to authenticate against the remote object.
    """
    yield
    remote.remove_menu("testmenu0", token)


@pytest.mark.usefixtures("cobbler_xmlrpc_base")
class TestMenu:
    @pytest.mark.usefixtures("create_testmenu", "remove_testmenu")
    def test_create_submenu(self, remote, token):
        """
        Test: create/edit a submenu object
        """
        # Arrange
        menus = remote.get_menus(token)

        # Act
        submenu = remote.new_menu(token)

        # Assert
        assert remote.modify_menu(submenu, "name", "testsubmenu0", token)
        assert remote.modify_menu(submenu, "parent", "testmenu0", token)

        assert remote.save_menu(submenu, token)

        new_menus = remote.get_menus(token)
        assert len(new_menus) == len(menus) + 1
        remote.remove_menu("testsubmenu0", token, False)

    @pytest.mark.usefixtures("remove_menu")
    def test_create_menu(self, remote, token):
        """
        Test: create/edit a menu object
        """
        # Arrange --> Nothing to arrange

        # Act & Assert
        menu = remote.new_menu(token)
        assert remote.modify_menu(menu, "name", "testmenu0", token)
        assert remote.modify_menu(menu, "display_name", "Test Menu0", token)
        assert remote.save_menu(menu, token)

    def test_get_menus(self, remote):
        """
        Test: Get menus
        """
        # Arrange --> Nothing to do

        # Act
        result = remote.get_menus()

        # Assert
        assert result == []

    @pytest.mark.usefixtures("create_menu", "remove_menu")
    def test_get_menu(self, remote, token):
        """
        Test: Get a menu object
        """

        # Arrange --> Done in fixture

        # Act
        menu = remote.get_menu("testmenu0")

        # Assert
        assert menu.get("name") == "testmenu0"

    @pytest.mark.usefixtures("create_menu", "remove_menu")
    def test_find_menu(self, remote, token):
        """
        Test: find a menu object
        """

        # Arrange --> Done in fixture

        # Act
        result = remote.find_menu({"name": "testmenu0"}, token)

        # Assert
        assert result

    @pytest.mark.usefixtures("create_menu", "remove_menu")
    def test_copy_menu(self, remote, token):
        """
        Test: copy a menu object
        """

        # Arrange --> Done in fixture

        # Act
        menu = remote.get_item_handle("menu", "testmenu0", token)

        # Assert
        assert remote.copy_menu(menu, "testmenucopy", token)

        # Cleanup
        remote.remove_menu("testmenucopy", token)

    @pytest.mark.usefixtures("create_menu")
    def test_rename_menu(self, remote, token):
        """
        Test: rename a menu object
        """

        # Arrange

        # Act
        menu = remote.get_item_handle("menu", "testmenu0", token)
        result = remote.rename_menu(menu, "testmenu1", token)

        # Assert
        assert result

        # Cleanup
        remote.remove_menu("testmenu1", token)

    @pytest.mark.usefixtures("create_menu")
    def test_remove_menu(self, remote, token):
        """
        Test: remove a menu object
        """

        # Arrange --> Done in fixture

        # Act
        result = remote.remove_menu("testmenu0", token)

        # Assert
        assert result



import logging
import os
import shutil
from contextlib import contextmanager
from pathlib import Path

import pytest


def pytest_addoption(parser):
    parser.addoption("-E", action="store", metavar="NAME", help="only run tests matching the environment NAME.")


def pytest_configure(config):
    # register an additional marker
    config.addinivalue_line("markers", "env(name): mark test to run only on named environment")


@contextmanager
def does_not_raise():
    yield


@pytest.fixture(scope="function")
def create_testfile(tmp_path):
    def _create_testfile(filename):
        path = os.path.join(tmp_path, filename)
        if not os.path.exists(path):
            Path(path).touch()
        return path
    return _create_testfile


@pytest.fixture(scope="function")
def create_kernel_initrd(create_testfile):
    def _create_kernel_initrd(name_kernel, name_initrd):
        create_testfile(name_kernel)
        return os.path.dirname(create_testfile(name_initrd))
    return _create_kernel_initrd


@pytest.fixture(scope="session", autouse=True)
def cleanup_leftover_items():
    """
    Will delete all JSON files which are left in Cobbler before a testrun!
    """
    logger = logging.getLogger("session-cleanup")
    cobbler_collections = ["distros", "files", "images", "menus", "mgmtclasses", "packages", "profiles", "repos",
                           "systems"]
    for collection in cobbler_collections:
        path = os.path.join("/var/lib/cobbler/collections", collection)
        for file in os.listdir(path):
            json_file = os.path.join(path, file)
            os.remove(json_file)
            logger.info("Removed file: " + json_file)


@pytest.fixture(scope="function")
def fk_initrd():
    """
    The path to the first fake initrd.

    :return: A filename as a string.
    """
    return "initrd1.img"


@pytest.fixture(scope="function")
def fk_initrd2():
    """
    The path to the second fake initrd.

    :return: A filename as a string.
    """
    return "initrd2.img"


@pytest.fixture(scope="function")
def fk_initrd3():
    """
    The path to the third fake initrd.

    :return: A path as a string.
    """
    return "initrd3.img"


@pytest.fixture(scope="function")
def fk_kernel():
    """
    The path to the first fake kernel.

    :return: A path as a string.
    """
    return "vmlinuz1"


@pytest.fixture(scope="function")
def fk_kernel2():
    """
    The path to the second fake kernel.

    :return: A path as a string.
    """
    return "vmlinuz2"


@pytest.fixture(scope="function")
def fk_kernel3():
    """
    The path to the third fake kernel.

    :return: A path as a string.
    """
    return "vmlinuz3"


@pytest.fixture(scope="function")
def redhat_autoinstall():
    """
    The path to the test.ks file for redhat autoinstall.

    :return: A path as a string.
    """
    return "test.ks"


@pytest.fixture(scope="function")
def suse_autoyast():
    """
    The path to the suse autoyast xml-file.
    :return: A path as a string.
    """
    return "test.xml"


@pytest.fixture(scope="function")
def ubuntu_preseed():
    """
    The path to the ubuntu preseed file.
    :return: A path as a string.
    """
    return "test.seed"
